@@ -27,11 +27,21 @@
 # define DEBUG_FILE "slrn.log" 
 #endif
 
-#include <config.h>
+#include "config.h"
+#include "features.h"
+
 
 #include <stdio.h>
 #include <string.h>
 
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+
+#ifdef HAVE_STDLIB_H
+# include <stdlib.h>
+#endif
+
 #ifdef VMS
 # include <signal.h>
 # ifdef MULTINET
@@ -66,10 +76,12 @@
 # include	<tiuser.h>
 # include	<stropts.h>
 # include	<sys/socket.h>
+# define SLRN_LOADED_SYS_SOCKET_H
 # ifdef WIN_TCP
 #  include	<sys/in.h>
 # else
 #  include	<netinet/in.h>
+#  define SLRN_LOADED_NETINET_IN_H
 # endif
 # define	IPPORT_NNTP	((unsigned short) 119)
 # include 	<netdb.h>	/* All TLI implementations may not have this */
@@ -79,6 +91,8 @@
 #   include "multinet_root:[multinet.include]netdb.h"
 #   include "multinet_root:[multinet.include.sys]socket.h"
 #   include "multinet_root:[multinet.include.netinet]in.h"
+#   define SLRN_LOADED_NETINET_IN_H
+#   define SLRN_LOADED_SYS_SOCKET_H
 #  endif
 #  ifdef NETLIB
 #   include <descrip.h>
@@ -92,36 +106,65 @@
 # else /* !VMS */
 #  include <sys/socket.h>
 #  include <netinet/in.h>
+#  define SLRN_LOADED_NETINET_IN_H
+#  define SLRN_LOADED_SYS_SOCKET_H
 #  ifndef EXCELAN
 #   include <netdb.h>
 #  endif /* !EXCELAN */
 # endif /* !VMS */
 #endif /* !TLI */
  
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-
-#ifdef HAVE_STDLIB_H
-# include <stdlib.h>
-#endif
 
 #include <sys/types.h>
 
 #ifdef HAVE_SYS_SOCKET_H
-# include <sys/socket.h>
+# ifndef SLRN_LOADED_SYS_SOCKET_H
+#  include <sys/socket.h>
+#  define SLRN_LOADED_SYS_SOCKET_H
+# endif
 #endif
 
 #ifdef HAVE_NETINET_IN_H
-# include <netinet/in.h>
+# ifndef SLRN_LOADED_NETINET_IN_H
+#  include <netinet/in.h>
+#  define SLRN_LOADED_NETINET_IN_H
+# endif
 #endif
 
 #ifdef HAVE_ARPA_INET_H
-# include <arpa/inet.h>
+# ifndef SLRN_LOADED_ARPA_INET_H
+#  include <arpa/inet.h>
+#  define SLRN_LOADED_ARPA_INET_H
+# endif
 #endif
 
-#define CLIENTLIB_C
-#include "clientlib.h"
+#ifdef VMS
+ 
+#  ifdef MULTINET
+#   define netread   socket_read
+#   define sock_write  socket_write
+#   define netclose  socket_close
+#  else
+#   if NETLIB
+#    define netread   netlib_read
+#    define sock_write  netlib_write
+#    define netclose  netlib_close
+#   else
+#    define netread   read
+#    define sock_write  write
+#    define netclose  close
+#   endif
+#  endif
+# else
+# endif /* VMS */
+
+#ifndef NETLIB
+# define SERVER_READ_FP  ser_rd_fp
+# define SERVER_WRITE_FP ser_wr_fp
+#else
+# define SERVER_READ_FP	ipctx
+# define SERVER_WRITE_FP	ipctx
+#endif
 
 #define HAS_BASIC_MEM_STUFF
 #ifdef HAS_BASIC_MEM_STUFF
@@ -149,15 +192,13 @@ int socket( int, struct sockproto *, struct sockaddr_in *, int );
 # include <netdnet/dnetdb.h>
 #endif /* DECNET */
 
-#include "nntp.h"
+#include "server.h"
 #include "misc.h"
 
 #ifndef NNTP_PORT
 # define NNTP_PORT 119
 #endif
 
-int Slrn_NNTP_Port = -1;
-
 
 #ifdef NEEDS_EXTERN_DECLARATIONS
 # define DECLARE_EXTERN(x) extern x;
@@ -168,13 +209,13 @@ int Slrn_NNTP_Port = -1;
 
 #ifdef VMS
 # ifdef NETLIB
-void *ipctx;
+static void *ipctx;
 # else
-int ser_rd_fp = -1, ser_wr_fp = -1;
+static int ser_rd_fp = -1, ser_wr_fp = -1;
 # endif
 #else
-FILE	*ser_rd_fp = NULL;
-FILE	*ser_wr_fp = NULL;
+static FILE	*ser_rd_fp = NULL;
+static FILE	*ser_wr_fp = NULL;
 #endif
 
 
@@ -182,6 +223,12 @@ FILE	*ser_wr_fp = NULL;
 static FILE *Debug_Fp;
 #endif
 
+static int Server_Has_Been_Closed = 1;
+
+
+static void client_close_server(void);
+static int client_get_server(char *, int);
+
 #ifdef VMS
 static void cancel_read (int sig_unused)
 {
@@ -191,64 +238,6 @@ static void cancel_read (int sig_unused)
 
 static int get_tcp_socket(char *);
 
-
-/*
- * getserverbyfile	Get the name of a server from a named file.
- *			Handle white space and comments.
- *			Use NNTPSERVER environment variable if set.
- *
- *	Parameters:	"file" is the name of the file to read.
- *
- *	Returns:	Pointer to static data area containing the
- *			first non-ws/comment line in the file.
- *			NULL on error (or lack of entry in file).
- *
- *	Side effects:	None.
- */
-
-char *getserverbyfile(char	*file)
-{
-   register FILE	*fp;
-   register char	*cp;
-   static char	buf[256];
-   
-   if (NULL != (cp = getenv("NNTPSERVER"))) 
-     {
-	(void)
-       strcpy(buf, cp);
-	return (buf);
-     }
-   
-   if (file == NULL)
-     {
-#ifdef NNTPSERVER_NAME
-	strcpy (buf, NNTPSERVER_NAME);
-	return buf;
-#else
-	return (NULL);
-#endif
-     }
-   
-   fp = fopen(file, "r");
-   if (fp == NULL)
-     return (NULL);
-   
-   while (fgets(buf, sizeof (buf), fp) != NULL) 
-     {
-	if (*buf == '\n' || *buf == '#')
-	  continue;
-	cp = slrn_strchr(buf, '\n');
-	if (cp)
-	  *cp = '\0';
-	(void) fclose(fp);
-	return (buf);
-     }
-   
-   (void) fclose(fp);
-   return (NULL);			 /* No entry */
-}
-
-
 /*
  * client_server_init  Get a connection to the remote news server.
  *
@@ -262,12 +251,15 @@ char *getserverbyfile(char	*file)
  *			for reading and writing to server.
  */
 
-int client_server_init(char *machine, char *line, int len)
+static int client_server_init(char *machine, char *line, int len)
 {
    int	sockt_rd, sockt_wr;
 #ifdef DECNET
    char	*cp;
    
+   if (Server_Has_Been_Closed == 0) 
+     client_close_server ();
+   
    cp = slrn_strchr(machine, ':');
    
    if (cp && cp[1] == ':') 
@@ -278,6 +270,8 @@ int client_server_init(char *machine, char *line, int len)
    else
      sockt_rd = get_tcp_socket(machine);
 #else
+   if (Server_Has_Been_Closed == 0) 
+     client_close_server ();
    sockt_rd = get_tcp_socket(machine);
 #endif
    
@@ -323,7 +317,10 @@ int client_server_init(char *machine, char *line, int len)
 #ifdef DEBUG_FILE
    Debug_Fp = fopen (DEBUG_FILE, "w");
 #endif
+   
+   Server_Has_Been_Closed = 0;
    (void) client_get_server(line, len);
+   
    return (atoi(line));
 }
 
@@ -787,15 +784,25 @@ static int handle_server_response(int response, char *nntpserver)
  *			fflush.
  */
 
-int
+static int
 client_put_server(char *string)
 {
 #ifdef VMS
    char *cp;
+# ifdef NETLIB
+   int rc;
+   struct dsc$descriptor dsc;
+   unsigned short retlen;
+# endif
 #endif
+   
+   if (Server_Has_Been_Closed)
+     return -1;
+   
 #ifdef DEBUG_FILE
    if (Debug_Fp != NULL) fprintf(Debug_Fp, ">>> %s\n", string);
 #endif
+   
 #ifdef VMS
 # ifndef NETLIB
    cp = (char *)malloc(strlen(string)+4);
@@ -804,18 +811,13 @@ client_put_server(char *string)
    /* sock_write(ser_wr_fp, cp, strlen(cp)); */
    free(cp);
 # else
-   int rc;
-   struct dsc$descriptor dsc;
-   unsigned short retlen;
-   
    INIT_SDESC(dsc, strlen(string), string);
-   
    rc = netlib_writeline(&ipctx, &dsc);
-   if (!(rc & 1))
-     return(-1);
+   if (!(rc & 1)) return(-1);
 # endif /* !NETLIB */
 #else
-   if ((EOF == fprintf(ser_wr_fp, "%s\r\n", string))
+   if ((EOF == fputs (string, ser_wr_fp))
+       || (EOF == fputs ("\r\n", ser_wr_fp))
        || (EOF == fflush(ser_wr_fp)))
      return -1;
 #endif
@@ -842,7 +844,7 @@ client_put_server(char *string)
 static char buffer[BUFSIZE];
 static int  bufread = 0;
 
-int client_get_server(char *string, int size)
+static int client_get_server(char *string, int size)
 {
    register char *cp;
    int status, nb, len, count;
@@ -854,6 +856,8 @@ int client_get_server(char *string, int size)
      }
    r_iosb;
    
+   if (Server_Has_Been_Closed) return -1;
+   
    if (!bufread)
      {
 	signal(SIGALRM,cancel_read);
@@ -912,7 +916,7 @@ int client_get_server(char *string, int size)
 #else /* !VMS || NETLIB */
 # ifdef NETLIB
 #  define TIMEOUT 60
-int client_get_server(char *string, int size)
+static int client_get_server(char *string, int size)
 {
    register char *cp;
    int rc;
@@ -920,6 +924,7 @@ int client_get_server(char *string, int size)
    unsigned short retlen;
    
    INIT_SDESC(dsc, size, string);
+   if (Server_Has_Been_Closed) return -1;
    rc = netlib_readline(&ipctx, &dsc, &retlen, 0, 0);
    if (!(rc & 1))
      return (-1);
@@ -933,10 +938,12 @@ int client_get_server(char *string, int size)
 }
 
 # else /* !VMS && !NETLIB */
-int client_get_server(char *string, int size)
+static int client_get_server(char *string, int size)
 {
    register char *cp;
    
+   if (Server_Has_Been_Closed) return -1;
+
    if (fgets(string, size, ser_rd_fp) == NULL)
      return (-1);
    
@@ -967,10 +974,13 @@ int client_get_server(char *string, int size)
  *			after this routine is called.
  */
 
-void
+static void
 client_close_server(void)
 {
    char	ser_line[256];
+   
+   if (Server_Has_Been_Closed) return;
+
 #ifdef VMS
 # ifndef NETLIB
    if (ser_wr_fp <= 0 || ser_rd_fp <= 0) return;
@@ -996,11 +1006,12 @@ client_close_server(void)
    (void) fclose(ser_rd_fp);
 # endif
 #endif
+   Server_Has_Been_Closed = 1;
 }
 
 #ifdef VMS
 # ifdef NETLIB
-int netwrite(void *sockt, char *string, int len)
+static int netwrite(void *sockt, char *string, int len)
 {
    struct dsc$descriptor dsc;
    int rc;
@@ -1017,7 +1028,7 @@ int netwrite(void *sockt, char *string, int len)
 }
 # else
 
-int netwrite(int sockt, char *string, int len)
+static int netwrite(int sockt, char *string, int len)
 {
 #  ifdef DEBUG_FILE
    if (Debug_Fp != NULL) fprintf(Debug_Fp, ">>> %s\n", string);
@@ -1026,4 +1037,3 @@ int netwrite(int sockt, char *string, int len)
 }
 # endif  /* !NETLIB */
 #endif  /* VMS */
-
