@@ -1,7 +1,8 @@
 /*  Copyright (c) 1995 John E. Davis (davis@space.mit.edu)
  *  All rights reserved.
  */
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -27,7 +28,7 @@
 #include "art.h"
 #include "misc.h"
 #include "post.h"
-#include "nntp.h"
+#include "server.h"
 #include "hash.h"
 #include "score.h"
 #include "menu.h"
@@ -65,7 +66,11 @@ int Slrn_Query_Group_Cutoff = 100;
 int Slrn_Groups_Dirty;	       /* greater than 0 if need to write newsrc */
 int Slrn_List_Active_File = 0;
 int Slrn_Use_Xgtitle = 0;
+int Slrn_Write_Newsrc_Flags = 0;       /* if 1, do not save unsubscribed 
+					* if 2, do not save new unsubscribed.
+					*/
 
+int Slrn_Group_Description_Column = 40;/* column where group descr start */
 static Group_Type *Groups, *Current_Group;
 
 static Group_Type *Top_Group;		       /* group at top of screen */
@@ -75,7 +80,7 @@ static int Num_Groups;
 static int Line_Num;
 static int Groups_Hidden;	       /* if true, hide groups with no arts */
 
-int Group_Window_Size;		       /* num rows in group window */
+int Slrn_Group_Window_Size;		       /* num rows in group window */
 int Slrn_Group_Display_Descriptions = 1;
 int Slrn_No_Backups = 0;
 int Slrn_Prompt_Next_Group = 1;
@@ -120,7 +125,15 @@ void slrn_add_group_ranges (int min, int max)
     * the range of articles nolonger available.
     */
    next = &Current_Group->range;
-   if (max < next->min) max = next->min - 1;
+   if (max < next->min) 
+     {
+	/* If we have already expanded the range up to range currently available
+	 * at server and we are now trying to add another range below available
+	 * range, do not bother.
+	 */
+	if (next->next != NULL) return;
+	max = next->min - 1;
+     }
    
    /* Count number unread */
    unread = next->max;
@@ -204,6 +217,7 @@ static void group_mark_article_as_read (Group_Type *g, long num)
      {
 	slrn_exit_error ("Memory allocation error");
      }
+   
    newr->min = newr->max = num;
    newr->next = r;
    if (r != NULL) r->prev = newr;
@@ -231,10 +245,12 @@ void slrn_mark_article_as_read (char *group, long num)
      {
 	if ((g->hash == hash) && !strcmp (group, g->name))
 	  {
-	     if (g->flags & UNSUBSCRIBED) return;
+	     /* If it looks like we have read this group, mark it read. */
+	     if (((g->flags & UNSUBSCRIBED) == 0)
+		 || (g->range.next != NULL))
+	       group_mark_article_as_read (g, num);
 	     
-	     group_mark_article_as_read (g, num);
-	     return;
+	     break;
 	  }
 	g = g->hash_next;
      }
@@ -289,10 +305,15 @@ static Group_Type *create_group_entry (char *name, unsigned int len,
    
    if (query_server)
      {
-	if (0 != nntp_select_group (g->name, &min, &max))
+	if (0 != Slrn_Server_Obj->select_group (g->name, &min, &max))
 	  {
 	     Slrn_Groups_Dirty = 1;
-	     slrn_error ("Group %s is bogus.\r\n", g->name);
+	     
+	     if (Slrn_TT_Initialized == 0)
+	       slrn_error ("Group %s is bogus.\r\n", g->name);
+	     else
+	       slrn_error ("Group %s is bogus.\n", g->name);
+	     
 	     SLFREE (g);
 	     return NULL;
 	  }
@@ -319,6 +340,7 @@ static Group_Type *create_group_entry (char *name, unsigned int len,
      }
    else
      {
+	if (Current_Group == NULL) Current_Group = Groups;
 	g->next = Current_Group->next;
 	if (g->next != NULL) g->next->prev = g;
 	Current_Group->next = g;
@@ -535,7 +557,7 @@ static Group_Type *process_xgtitle_info (void)
    char buf [NNTP_BUFFER_SIZE];
    Group_Type *first = NULL, *save = Current_Group;
    
-   while (nntp_read_line(buf, sizeof(buf)) != NULL)
+   while (Slrn_Server_Obj->read_line(buf, sizeof(buf)) != NULL)
      {
 	char *b, ch;
 	unsigned int len;
@@ -621,7 +643,7 @@ static void toggle_list_all_groups1 (int hide_flag)
 
 	if ((Slrn_List_Active_File == 0)
 	    && Slrn_Use_Xgtitle
-	    && (-1 != nntp_xgtitle_cmd (pat)))
+	    && (-1 != Slrn_Server_Obj->xgtitle_cmd (pat)))
 	  {
 	     first_found = process_xgtitle_info ();
 	  }
@@ -684,7 +706,8 @@ static void toggle_list_all_groups1 (int hide_flag)
    if ((all_hidden == 0) && (Current_Group == NULL))
      {
 	Current_Group = Groups;
-	if (Current_Group->flags & HIDDEN)
+	if ((Current_Group != NULL)
+	    && (Current_Group->flags & HIDDEN))
 	  {
 	     Current_Group = NULL;
 	  }
@@ -800,9 +823,21 @@ int slrn_write_newsrc (void)
    static FILE *fp;
    int pass;
    int max;
+#ifdef unix
+# ifndef __os2__
+   struct stat filestat;
+   int stat_worked;
+# endif
+#endif
+
+   slrn_init_hangup_signals (0);
    
-   if (Slrn_Groups_Dirty == 0) return 0;
-   
+   if (Slrn_Groups_Dirty == 0) 
+     {
+	slrn_init_hangup_signals (1);
+	return 0;
+     }
+
    /* In case of hangup and we were writing the file, make sure it is closed.
     * This will not hurt since we are going to do it again anyway.
     */
@@ -812,6 +847,14 @@ int slrn_write_newsrc (void)
    slrn_smg_refresh ();
    
    slrn_make_home_filename (Slrn_Newsrc_File, file);
+   
+#ifdef unix
+# ifndef __os2__
+   /* Try to preserve .newsrc permissions and owner/group */
+   stat_worked = (-1 != stat (file, &filestat));
+# endif
+#endif
+  
    if (Slrn_No_Backups == 0)
      {
 #ifdef VMS
@@ -827,6 +870,7 @@ int slrn_write_newsrc (void)
      {
 	slrn_error ("Unable to save to file %s.", file);
 	if (Slrn_No_Backups == 0) (void) rename (backup_file, file);
+	slrn_init_hangup_signals (1);
 	return -1;
      }
    
@@ -851,12 +895,22 @@ int slrn_write_newsrc (void)
 	       }
 	     else if (pass == 1)
 	       {
-		  if (((g->flags & UNSUBSCRIBED) == 0)
-		      || (0 == (g->flags & GROUP_PROCESSED)))
+		  if ((g->flags & UNSUBSCRIBED) == 0)
 		    {
 		       g = g->next;
 		       continue;
 		    }
+		  
+		  if (Slrn_Write_Newsrc_Flags)
+		    {
+		       if ((Slrn_Write_Newsrc_Flags == 1)
+			   || ((Slrn_Write_Newsrc_Flags == 2)
+			       && (g->range.next == NULL)))
+			 {
+			    g = g->next;
+			    continue;
+			 }
+		    }
 		  fputs (g->name, fp);
 		  putc ('!', fp);
 	       }
@@ -894,12 +948,26 @@ int slrn_write_newsrc (void)
 	  }
      }
    slrn_fclose (fp); fp = NULL;
-#ifndef VMS
-   (void) chmod (file, 0600);
+   
+#ifdef unix
+# ifndef __os2__
+   /* Try to preserve .newsrc permissions and owner/group */
+   if (stat_worked)
+     {
+ 	if (-1 == chmod (file, filestat.st_mode & 0777))
+	  (void) chmod (file, 0600);
+
+ 	(void) chown (file, filestat.st_uid, filestat.st_gid);
+     }
+   else
+     (void) chmod (file, 0600 );
+# endif
 #endif
+
    Slrn_Groups_Dirty = 0;
    if (Slrn_TT_Initialized & 1)
      slrn_message ("Writing %s... done.", Slrn_Newsrc_File);
+   slrn_init_hangup_signals (1);
    return 0;
 }
 
@@ -928,6 +996,7 @@ static void free_all_groups (void)
    Groups = NULL;
    Current_Group = NULL;
    Top_Group = Bottom_Group = NULL;
+   Num_Groups = Line_Num = 0;
    for (i = 0; i < GROUP_HASH_TABLE_SIZE; i++) Group_Hash_Table[i] = NULL;
 }
 
@@ -946,6 +1015,11 @@ static void save_newsrc (void)
 
 static void refresh_groups (void)
 {
+   char name[MAX_GROUP_NAME + 1];
+   
+   *name = 0;
+   if (Current_Group != NULL) strcpy (name, Current_Group->name);
+   
    slrn_set_suspension (1);
    save_newsrc ();
    free_all_groups ();
@@ -955,7 +1029,12 @@ static void refresh_groups (void)
     * time.
     */
    Groups_Hidden = !Groups_Hidden;
-   slrn_get_new_news (1, 0);	       /* no new groups */
+   if (-1 == slrn_get_new_news (1, 0))
+     {
+	slrn_quit (0);
+     }
+   if (*name)
+     (void) find_group (name);
    slrn_set_suspension (0);
    quick_help ();
 }
@@ -1122,7 +1201,7 @@ static void subscribe (void)
    if (Slrn_Prefix_Arg_Ptr == NULL)
      {
 	Current_Group->flags &= ~UNSUBSCRIBED;
-	Current_Group->flags |= (TOUCHED | GROUP_PROCESSED);
+	Current_Group->flags |= TOUCHED;
 	group_down_n (1);
 	Slrn_Groups_Dirty = 1;
 	return;
@@ -1145,7 +1224,7 @@ static void subscribe (void)
 	       {
 		  g->flags &= ~HIDDEN;
 		  g->flags &= ~UNSUBSCRIBED;
-		  g->flags |= (TOUCHED | GROUP_PROCESSED);
+		  g->flags |= TOUCHED;
 	       }
 	  }
 	g = g->next;
@@ -1217,7 +1296,7 @@ static void unsubscribe (void)
 		 && (NULL != SLang_regexp_match ((unsigned char *) g->name, min_len, &pat_regexp)))
 	       {
 		  g->flags &= ~HIDDEN;
-		  g->flags |= (TOUCHED | GROUP_PROCESSED | UNSUBSCRIBED);
+		  g->flags |= (TOUCHED | UNSUBSCRIBED);
 	       }
 	  }
 	g = g->next;
@@ -1261,15 +1340,23 @@ static void digit_arg (void)
 
 static void group_select_group (void)
 {
-   int min, max, n;
+   int min, max, n, max_available;
    char *group;
    SLang_Key_Type *key;
    int ret;
    void (*f)(void);
    Slrn_Range_Type *r;
+   int prefix;
    
    top:
    
+   if (Slrn_Prefix_Arg_Ptr != NULL)
+     {
+	prefix = *Slrn_Prefix_Arg_Ptr;
+	Slrn_Prefix_Arg_Ptr = NULL;
+     }
+   else prefix = 0;
+   
    if (Current_Group == NULL)
      {
 	return;
@@ -1279,7 +1366,7 @@ static void group_select_group (void)
    
    slrn_message ("Selecting %s...", group);
    slrn_smg_refresh ();
-   if (nntp_select_group (group, &min, &max) < 0)
+   if (Slrn_Server_Obj->select_group (group, &min, &max) < 0)
      {
 	Current_Group->flags &= ~UNSUBSCRIBED;
 	slrn_error ("This group appears to be bogus.");
@@ -1288,10 +1375,19 @@ static void group_select_group (void)
    
    if (max == 0)
      {
+	int nmax, nmin;
+	
+	nmax = Current_Group->range.max;
+	nmin = Current_Group->range.min;
+	
+	/* Server database inconsistent. */
+	for (n = nmin; n <= nmax; n++)
+	  group_mark_article_as_read (Current_Group, n);
+	
+	/* Current_Group->unread = 0; */
+	slrn_message ("No articles to read.");
 	Slrn_Full_Screen_Update = 1;
 	Slrn_Groups_Dirty = 1;
-	Current_Group->unread = 0;
-	slrn_message ("No articles to read.");
 	return;
      }
    
@@ -1306,13 +1402,46 @@ static void group_select_group (void)
 	  group_mark_article_as_read (Current_Group, n);
      }
    else Current_Group->range.max = max;
+
+   /* In case more articles arrived at the server between the time that 
+    * slrn was first started and when the server was just queried, update
+    * the ranges of read/unread articles.
+    */
    
+   max_available = Current_Group->range.max - Current_Group->range.min + 1;
+   
+   r = &Current_Group->range;
+   if (r->next != NULL)
+     {
+	if (r->next->min <= r->min)
+	  r->next->min = 1;
+	
+	n = r->max;
+	while (r->next != NULL)
+	  {
+	     r = r->next;
+	     n -= r->max - r->min + 1;
+	  }
+	if (n < 0) n = 0;
+	if (n > max_available) n = max_available;
+	Current_Group->unread = n;
+     }
+   
+   n = Current_Group->unread;
+
+#if 1
+   if ((prefix & 1) || (n == 0))
+     n = max_available;
+#else
+
    if ((Slrn_Prefix_Arg_Ptr == NULL)
        && (Current_Group->unread > 0)) n = Current_Group->unread;
    else n = max - min + 1;
+#endif
    
-   if ((n > Slrn_Query_Group_Cutoff)
-       && (Slrn_Query_Group_Cutoff > 0))
+   if ((prefix & 1) || 
+       ((n > Slrn_Query_Group_Cutoff)
+	&& (Slrn_Query_Group_Cutoff > 0)))
      {
 	char int_prompt_buf[256];
 	sprintf (int_prompt_buf, "%s: Read how many?", Current_Group->name);
@@ -1324,7 +1453,7 @@ static void group_select_group (void)
 	     return;
 	  }
 	
-	if ((Slrn_Prefix_Arg_Ptr == NULL)
+	if ((0 == (prefix & 1))
 	    && (Current_Group->unread != 0))
 	  {
 	     r = Current_Group->range.next;
@@ -1332,16 +1461,21 @@ static void group_select_group (void)
 	       {
 		  while (r->next != NULL) r = r->next;
 		  if (r->max + n > max)
-		    n = -n;
+		    n = -n;	       /* special treatment in article mode
+					* because we will need to query the 
+					* server about articles in a group 
+					* that we have already read.
+					*/
 	       }
 	  }
      }
-   else if ((Slrn_Prefix_Arg_Ptr == NULL) && (Current_Group->unread != 0)) n = 0;
-   
-   Slrn_Prefix_Arg_Ptr = NULL;
+   else if ((0 == (prefix & 1)) && (Current_Group->unread != 0))
+     n = 0;
    
    Slrn_redraw_function = NULL;
-   ret = slrn_article_mode (Current_Group->name, &Current_Group->range, n);
+   
+   ret = slrn_article_mode (Current_Group->name, &Current_Group->range, 
+			    n, ((prefix & 2) == 0));
    Slrn_redraw_function = group_update_screen;
    Slrn_Winch_Function = NULL;
    
@@ -1366,7 +1500,8 @@ static void group_select_group (void)
 			    slrn_smg_refresh ();
 			    
 			    key = SLang_do_key (Slrn_Group_Keymap, (int (*)(void)) SLang_getkey);
-			    if (key == NULL) f = NULL; else f = (void (*)(void)) key->f;
+			    if (key == NULL) f = NULL; 
+			    else f = (void (*)(void)) key->f.f;
 			 }
 		       
 		       if (f == group_select_group)  goto top;
@@ -1406,7 +1541,7 @@ void slrn_post_cmd (void)
    char group[256];
    char subj[256];
    
-   if (Slrn_Can_Post == 0)
+   if (Slrn_Post_Obj->can_post == 0)
      {
 	slrn_error ("Posting not allowed.");
 	return;
@@ -1468,14 +1603,14 @@ static void group_pageup (void)
 	     Line_Num -= n;
 	     n = Line_Num;	       /* save this */
 	     /* Now compute new top group */
-	     group_up_n (Group_Window_Size - 1);
+	     group_up_n (Slrn_Group_Window_Size - 1);
 	     Top_Group = Current_Group;
 	     Current_Group = g; Line_Num = n;
 	     return;
 	  }
      }
    
-   if (0 == group_up_n (Group_Window_Size - 1))
+   if (0 == group_up_n (Slrn_Group_Window_Size - 1))
      slrn_error ("Top of Buffer.");
 }
 
@@ -1503,7 +1638,7 @@ static void group_pagedown (void)
 	  }
      }
    
-   if (0 == group_down_n (Group_Window_Size - 1))
+   if (0 == group_down_n (Slrn_Group_Window_Size - 1))
      slrn_error ("End of Buffer.");
 }
 
@@ -1683,77 +1818,77 @@ void slrn_init_group_keymap (void)
    
    Slrn_Group_Keymap->functions = Group_Functions;
    
-   SLang_define_key1("\0331", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0332", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0333", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0334", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0335", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0336", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0337", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0338", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0339", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1("\0330", (VOID_STAR) digit_arg, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^K\033[A", (VOID_STAR) group_bob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^K\033OA", (VOID_STAR) group_bob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^K\033[B", (VOID_STAR) group_eob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^K\033OB", (VOID_STAR) group_eob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033A", (VOID_STAR) toggle_group_display, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033>", (VOID_STAR) group_eob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033<", (VOID_STAR) group_bob, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^D", (VOID_STAR) group_pagedown, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^V", (VOID_STAR) group_pagedown, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key ("\0331", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0332", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0333", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0334", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0335", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0336", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0337", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0338", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0339", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key ("\0330", (FVOID_STAR) digit_arg, Slrn_Group_Keymap);
+   SLkm_define_key  ("^K\033[A", (FVOID_STAR) group_bob, Slrn_Group_Keymap);
+   SLkm_define_key  ("^K\033OA", (FVOID_STAR) group_bob, Slrn_Group_Keymap);
+   SLkm_define_key  ("^K\033[B", (FVOID_STAR) group_eob, Slrn_Group_Keymap);
+   SLkm_define_key  ("^K\033OB", (FVOID_STAR) group_eob, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033A", (FVOID_STAR) toggle_group_display, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033>", (FVOID_STAR) group_eob, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033<", (FVOID_STAR) group_bob, Slrn_Group_Keymap);
+   SLkm_define_key  ("^D", (FVOID_STAR) group_pagedown, Slrn_Group_Keymap);
+   SLkm_define_key  ("^V", (FVOID_STAR) group_pagedown, Slrn_Group_Keymap);
 #ifdef __os2__
-   SLang_define_key1 ("^@Q", (VOID_STAR) group_pagedown, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\xE0Q", (VOID_STAR) group_pagedown, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^@I", (VOID_STAR) group_pageup, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\xE0I", (VOID_STAR) group_pageup, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);   
+   SLkm_define_key  ("^@Q", (FVOID_STAR) group_pagedown, Slrn_Group_Keymap);
+   SLkm_define_key  ("\xE0Q", (FVOID_STAR) group_pagedown, Slrn_Group_Keymap);
+   SLkm_define_key  ("^@I", (FVOID_STAR) group_pageup, Slrn_Group_Keymap);
+   SLkm_define_key  ("\xE0I", (FVOID_STAR) group_pageup, Slrn_Group_Keymap);   
 #else
-   SLang_define_key1 ("\033[6~", (VOID_STAR) group_pagedown, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033[5~", (VOID_STAR) group_pageup, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[6~", (FVOID_STAR) group_pagedown, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[5~", (FVOID_STAR) group_pageup, Slrn_Group_Keymap);
 #endif
-   SLang_define_key1 ("^U", (VOID_STAR) group_pageup, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033V", (VOID_STAR) group_pageup, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("a", (VOID_STAR) add_group_cmd, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("u", (VOID_STAR) unsubscribe, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("s", (VOID_STAR) subscribe, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033u", (VOID_STAR) uncatch_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("c", (VOID_STAR) catch_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("K", (VOID_STAR) toggle_scoring, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("L", (VOID_STAR) toggle_list_all_groups, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("l", (VOID_STAR) toggle_hide_groups, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^Z", (VOID_STAR) slrn_suspend_cmd, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 (" ", (VOID_STAR) group_select_group, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 (".", (VOID_STAR) slrn_repeat_last_key, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("P", (VOID_STAR) slrn_post_cmd, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("?", (VOID_STAR) slrn_group_help, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\r", (VOID_STAR) group_select_group, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("q", (VOID_STAR) group_quit, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^X^C", (VOID_STAR) group_quit, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^X^T", (VOID_STAR) transpose_groups, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^R", (VOID_STAR) slrn_redraw, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^L", (VOID_STAR) slrn_redraw, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^P", (VOID_STAR) group_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("^U", (FVOID_STAR) group_pageup, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033V", (FVOID_STAR) group_pageup, Slrn_Group_Keymap);
+   SLkm_define_key  ("a", (FVOID_STAR) add_group_cmd, Slrn_Group_Keymap);
+   SLkm_define_key  ("u", (FVOID_STAR) unsubscribe, Slrn_Group_Keymap);
+   SLkm_define_key  ("s", (FVOID_STAR) subscribe, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033u", (FVOID_STAR) uncatch_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("c", (FVOID_STAR) catch_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("K", (FVOID_STAR) toggle_scoring, Slrn_Group_Keymap);
+   SLkm_define_key  ("L", (FVOID_STAR) toggle_list_all_groups, Slrn_Group_Keymap);
+   SLkm_define_key  ("l", (FVOID_STAR) toggle_hide_groups, Slrn_Group_Keymap);
+   SLkm_define_key  ("^Z", (FVOID_STAR) slrn_suspend_cmd, Slrn_Group_Keymap);
+   SLkm_define_key  (" ", (FVOID_STAR) group_select_group, Slrn_Group_Keymap);
+   SLkm_define_key  (".", (FVOID_STAR) slrn_repeat_last_key, Slrn_Group_Keymap);
+   SLkm_define_key  ("P", (FVOID_STAR) slrn_post_cmd, Slrn_Group_Keymap);
+   SLkm_define_key  ("?", (FVOID_STAR) slrn_group_help, Slrn_Group_Keymap);
+   SLkm_define_key  ("\r", (FVOID_STAR) group_select_group, Slrn_Group_Keymap);
+   SLkm_define_key  ("q", (FVOID_STAR) group_quit, Slrn_Group_Keymap);
+   SLkm_define_key  ("^X^C", (FVOID_STAR) group_quit, Slrn_Group_Keymap);
+   SLkm_define_key  ("^X^T", (FVOID_STAR) transpose_groups, Slrn_Group_Keymap);
+   SLkm_define_key  ("^R", (FVOID_STAR) slrn_redraw, Slrn_Group_Keymap);
+   SLkm_define_key  ("^L", (FVOID_STAR) slrn_redraw, Slrn_Group_Keymap);
+   SLkm_define_key  ("^P", (FVOID_STAR) group_up, Slrn_Group_Keymap);
 #ifdef __os2__
-   SLang_define_key1 ("^@H", (VOID_STAR) group_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\xE0H", (VOID_STAR) group_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^@P", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\xE0P", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("^@H", (FVOID_STAR) group_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("\xE0H", (FVOID_STAR) group_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("^@P", (FVOID_STAR) group_down, Slrn_Group_Keymap);
+   SLkm_define_key  ("\xE0P", (FVOID_STAR) group_down, Slrn_Group_Keymap);
 #else
-   SLang_define_key1 ("\033[A", (VOID_STAR) group_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033OA", (VOID_STAR) group_up, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033[B", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033OB", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[A", (FVOID_STAR) group_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033OA", (FVOID_STAR) group_up, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[B", (FVOID_STAR) group_down, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033OB", (FVOID_STAR) group_down, Slrn_Group_Keymap);
 #endif
-   SLang_define_key1 ("N", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("^N", (VOID_STAR) group_down, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("/", (VOID_STAR) group_search, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("G", (VOID_STAR) refresh_groups, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("X", (VOID_STAR) save_newsrc, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("N", (FVOID_STAR) group_down, Slrn_Group_Keymap);
+   SLkm_define_key  ("^N", (FVOID_STAR) group_down, Slrn_Group_Keymap);
+   SLkm_define_key  ("/", (FVOID_STAR) group_search, Slrn_Group_Keymap);
+   SLkm_define_key  ("G", (FVOID_STAR) refresh_groups, Slrn_Group_Keymap);
+   SLkm_define_key  ("X", (FVOID_STAR) save_newsrc, Slrn_Group_Keymap);
    
    /* mouse (left/right/middle) */
-   SLang_define_key1 ("\033[M\040", (VOID_STAR) group_mouse_left, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033[M\041", (VOID_STAR) group_mouse_middle, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
-   SLang_define_key1 ("\033[M\042", (VOID_STAR) group_mouse_right, SLKEY_F_INTRINSIC, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[M\040", (FVOID_STAR) group_mouse_left, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[M\041", (FVOID_STAR) group_mouse_middle, Slrn_Group_Keymap);
+   SLkm_define_key  ("\033[M\042", (FVOID_STAR) group_mouse_right, Slrn_Group_Keymap);
    
    if (SLang_Error) slrn_exit_error (err);
 }
@@ -1766,16 +1901,16 @@ void slrn_do_key (SLKeyMap_List_Type *map)
    static SLang_Key_Type *last_key;
    
    key = SLang_do_key (map, (int (*)(void)) SLang_getkey);
-   if (Slrn_Message_Present) slrn_clear_message ();
+   if (Slrn_Message_Present || SLang_Error) slrn_clear_message ();
    SLang_Error = SLKeyBoard_Quit = 0;
    
-   if ((key == NULL) || (key->f == NULL))
+   if ((key == NULL) || (key->f.f == NULL))
      {
 	SLtt_beep ();
      }
    else
      {
-	if ((map == last_map) && (key->f == (VOID_STAR) slrn_repeat_last_key))
+	if ((map == last_map) && (key->f.f == (FVOID_STAR) slrn_repeat_last_key))
 	  key = last_key;
 	
 	/* set now to avoid problems with recursive call */
@@ -1783,7 +1918,7 @@ void slrn_do_key (SLKeyMap_List_Type *map)
 	last_map = map;
 	
 	if (key->type == SLKEY_F_INTRINSIC)
-	  (((void (*)(void))(key->f)) ());
+	  (((void (*)(void))(key->f.f)) ());
 	else
 	  {
 	     SLtt_beep ();
@@ -1829,7 +1964,11 @@ void slrn_get_group_descriptions (void)
 #ifdef VMS
    sprintf (file, "%s-dsc", Slrn_Newsrc_File);
 #else
+# ifdef __os2__
+   sprintf (file, "ds-%s", Slrn_Newsrc_File);
+# else
    sprintf (file, "%s.dsc", Slrn_Newsrc_File);
+# endif
 #endif
    
    
@@ -1840,14 +1979,14 @@ Unable to create newsgroup description file:\n%s\n", line);
      }
    
    fprintf (stdout, "Creating description file %s.\n", line);
-   nntp_list_newsgroups ();
+   Slrn_Server_Obj->list_newsgroups ();
    fprintf (stdout, "Getting newsgroup descriptions from server.\n\
 Note: This step may take some time if you have a slow connection!!!\n");
    
    fflush (stdout);
    
    num = 0;
-   while (NULL != nntp_read_line (line, 511))
+   while (NULL != Slrn_Server_Obj->read_line (line, 511))
      {
 	unsigned char *b, *bmax, *dsc, ch;
 	
@@ -1893,7 +2032,11 @@ int slrn_read_group_descriptions (void)
 #ifdef VMS
    sprintf (file, "%s-dsc", Slrn_Newsrc_File);
 #else
+# ifdef __os2__
+   sprintf (file, "ds-%s", Slrn_Newsrc_File);
+# else
    sprintf (file, "%s.dsc", Slrn_Newsrc_File);
+# endif
 #endif
    
    
@@ -1909,7 +2052,7 @@ int slrn_read_group_descriptions (void)
 		  fp = slrn_open_home_file (file, "r", line, 0);
 	       }
 #else
-	     sprintf (file, "%s/newsgroups.dsc", SLRN_DSC_DIR);
+	     sprintf (file, "%s/newsgroups.dsc", Slrn_Lib_Dir);
 	     fp = slrn_open_home_file (file, "r", line, 0);
 #endif
 	  }
@@ -2012,17 +2155,23 @@ void slrn_check_new_groups (int create_flag)
 	
 	parse_error = 0;
 	
-	slrn_put_server (line);
-	slrn_get_server (line, 255, 1);
-	if (atoi (line) != OK_NEWGROUPS)
+	switch (Slrn_Server_Obj->put_server_cmd (line, line, sizeof (line)))
 	  {
+	   case OK_NEWGROUPS:
+	     break;
+	     
+	   case ERR_FAULT:
+	     slrn_message ("Server does not implement NEWGROUPS command.");
+	     return;
+	     
+	   default:
 	     slrn_message ("Server failed to return proper response to NEWGROUPS:\n%s\n",
 			   line);
 	     goto parse_error_label;
 	  }
 	
 	num = 0;
-	while (NULL != nntp_read_line (line, 255))
+	while (NULL != Slrn_Server_Obj->read_line (line, 255))
 	  {
 	     /* line contains new newsgroup name */
 	     add_unsubscribed_group (line);
@@ -2125,9 +2274,9 @@ int slrn_read_newsrc (int create_flag)
    if (create_flag || Slrn_List_Active_File)
      {
 	int count = 0;
-	(void) nntp_list_active ();
+	(void) Slrn_Server_Obj->list_active ();
 	
-	while (NULL != nntp_read_line (line, sizeof(line)))
+	while (NULL != Slrn_Server_Obj->read_line (line, sizeof(line)))
 	  {
 	     unsigned int len;
 	     int min, max;
@@ -2252,10 +2401,16 @@ int slrn_read_newsrc (int create_flag)
    return 0;
 }
 
+static void slrn_group_hup (int sig)
+{
+   slrn_write_newsrc ();
+   slrn_quit (sig);
+}
+
 void slrn_group_mode (void)
 {
    quick_help ();
-   Slrn_Hangup_Hook = (void (*)(void)) slrn_write_newsrc;
+   Slrn_Hangup_Hook = slrn_group_hup;
    Slrn_redraw_function = group_update_screen;
    while (1)
      {
@@ -2273,7 +2428,7 @@ void slrn_group_mode (void)
 static void group_update_screen (void)
 {
    Group_Type *g;
-   int height = Group_Window_Size;
+   int height = Slrn_Group_Window_Size;
    int row;
    int curr_row = 0;
    
@@ -2330,7 +2485,7 @@ static void group_update_screen (void)
 	     if (Slrn_Full_Screen_Update || (g->flags & TOUCHED))
 	       {
 		  SLsmg_gotorc (row + 1, 0);
-		  SLsmg_printf ("  %c %5d  ",
+		  SLsmg_printf ("  %c%5d  ",
 				((g->flags & UNSUBSCRIBED) ? 'U'
 				 : ((g->flags & NEW_GROUP_FLAG) ? 'N' : ' ')),
 				g->unread);
@@ -2342,8 +2497,17 @@ static void group_update_screen (void)
 		    {
 		       if (g->descript != NULL)
 			 {
-			    SLsmg_gotorc (row + 1, 45);
+			    int dsc_row;
+			    
+			    dsc_row = SLsmg_get_column ();
+			    
+			    if (dsc_row < Slrn_Group_Description_Column)
+			      dsc_row = Slrn_Group_Description_Column;
+			    else dsc_row++;
+			    SLsmg_gotorc (row + 1, dsc_row);
+			    SLsmg_set_color (GROUP_DESCR_COLOR);
 			    SLsmg_write_string (g->descript);
+			    SLsmg_set_color (0);
 			 }
 		    }
 		  else if (g->range.max != -1)
