@@ -21,8 +21,9 @@ typedef struct Slrn_Range_Type
 } Slrn_Range_Type;
 
 extern int Slrn_Groups_Dirty;
-extern int Group_Window_Size;
+extern int Slrn_Group_Window_Size;
 extern int Slrn_Group_Display_Descriptions;
+extern int Slrn_Group_Description_Column;
 extern int *Slrn_Prefix_Arg_Ptr;
 extern int Slrn_No_Backups;
 extern int Slrn_Unsubscribe_New_Groups;
@@ -30,4 +31,4 @@ extern int Slrn_List_Active_File;
 extern int Slrn_Query_Group_Cutoff;
 extern int Slrn_Prompt_Next_Group;
 extern int Slrn_Use_Xgtitle;
-
+extern int Slrn_Write_Newsrc_Flags;
