@@ -1,15 +1,16 @@
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <string.h>
 
 
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+# include <unistd.h>
 #endif
 
 #ifdef HAVE_STDLIB_H
-#include <stdlib.h>
+# include <stdlib.h>
 #endif
 
 #include <slang.h>
@@ -56,6 +57,8 @@ static Menu_Type *Current_Menu;
 
 static void update_menu (Menu_Type *m)
 {
+   int col;
+   
    Current_Menu = m;
    /* if (Slrn_Full_Screen_Update == 0) return; */
    SLsmg_gotorc (0, 0);
@@ -66,7 +69,16 @@ static void update_menu (Menu_Type *m)
 	SLsmg_write_string ("   ");
 	m++;
      }
+   
    SLsmg_erase_eol ();
+
+   col = SLtt_Screen_Cols - 16;
+   if (SLsmg_get_column () < col)
+     SLsmg_gotorc (0, col);
+   
+   SLsmg_write_string ("slrn ");
+   SLsmg_write_string (Slrn_Version);
+   
    slrn_set_color (0);
 }
 
