@@ -4,7 +4,8 @@
  * Modified by John E. Davis <davis@space.mit.edu>
  */
 
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -21,14 +22,14 @@
 #include <ctype.h>
 
 #include <slang.h>
-#include "nntp.h"
+#include "server.h"
 #include "slrn.h"
 #include "misc.h"
 #include "slrn.h"
 #include "group.h"
 #include "art.h"
 
-#ifdef HAS_MIME_SUPPORT
+#if SLRN_HAS_MIME
 /* rest of file in this ifdef */
 
 #include "mime.h"
@@ -185,11 +186,23 @@ static int parse_content_type_line (void)
 	     b = slrn_skip_whitespace (b + 1);
 	     
 	     if (0 != slrn_case_strncmp ((unsigned char *)b,
-					 (unsigned char *)"charset=",
+					 (unsigned char *)"charset",
 					 8))
 	       continue;
 	     
-	     b += 8;
+	     b = slrn_skip_whitespace (b + 8);
+	     while (*b == 0)
+	       {
+		  line = line->next;
+		  if ((line == NULL)
+		      || ((line->flags & HEADER_LINE) == 0)
+		      || ((*(b = line->buf) != ' ') && (*b == '\t')))
+		    return -1;
+		  b = slrn_skip_whitespace (b);
+	       }
+	     
+	     if (*b != '=') continue;
+	     
 	     if (*b == '"') b++;
 	     charset = b;
 	     while (*b && (*b != ';')
@@ -499,7 +512,7 @@ void slrn_mime_process_article (void)
    Slrn_Mime_Was_Parsed = 1;	       /* or will be */
    
    rfc1522_decode_headers ();
-   
+
    if (NULL == find_header_line ("Mime-Version:")) return;
    if ((-1 == parse_content_type_line ())
        || (-1 == parse_content_transfer_encoding_line ()))
@@ -550,7 +563,7 @@ int slrn_mime_call_metamail (void)
        || slrn_get_yesno (1, "Process this MIME article with metamail") <= 0)
      return 0;
    
-   sprintf(tempfile, "/tmp/slrn.%d", getpid());
+   sprintf(tempfile, "/tmp/slrn.%ld", (long) getpid());
    
    fp = fopen(tempfile, "w");
    if (fp == NULL)
@@ -827,6 +840,9 @@ void slrn_mime_add_headers (FILE *fp)
 {
    char *encoding;
    
+   if (Mime_Posting_Charset == NULL)
+     Mime_Posting_Charset = "us-ascii";
+
    switch (Mime_Posting_Encoding)
      {
       default:
@@ -855,14 +871,12 @@ Content-Transfer-Encoding: %s\n",
      }
    else
      {
-	char buf[512];
-	sprintf (buf, "\
+	Slrn_Post_Obj->printf ("\
 Mime-Version: 1.0\r\n\
 Content-Type: text/plain; charset=%s\r\n\
 Content-Transfer-Encoding: %s\r\n",
 		 Mime_Posting_Charset,
 		 encoding);
-	nntp_puts(buf);
      }
 }
 
@@ -879,6 +893,6 @@ FILE *slrn_mime_encode (FILE *fp)
 
 
 
-#endif  /* HAS_MIME_SUPPORT */
+#endif  /* SLRN_HAS_MIME */
 
 
