@@ -1,7 +1,8 @@
 /*  Copyright (c) 1995 John E. Davis (davis@space.mit.edu)
  *  All rights reserved.
  */
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -32,8 +33,9 @@
 #include "group.h"
 #include "slrn.h"
 #include "post.h"
+#include "server.h"
 
-#ifdef HAS_MIME_SUPPORT
+#if SLRN_HAS_MIME
 #include "mime.h"
 #endif
 
@@ -56,12 +58,15 @@ char *Slrn_Editor;
 
 void slrn_set_color (int color)
 {
+#if 0
    if (SLtt_Use_Ansi_Colors == 0)
      {
 	if (color >= STATUS_COLOR) SLsmg_reverse_video ();
 	else SLsmg_normal_video ();
      }
-   else SLsmg_set_color (color);
+   else
+#endif
+     SLsmg_set_color (color);
 }
 
 void slrn_redraw (void)
@@ -74,7 +79,13 @@ int Slrn_Message_Present;
 void slrn_clear_message (void)
 {
    Slrn_Message_Present = Error_Present = 0;
-   SLang_Error = SLKeyBoard_Quit = 0;
+   /* SLang_Error = 0; */
+   
+   SLKeyBoard_Quit = 0;
+   
+   if ((Slrn_TT_Initialized & 1) == 0)
+     return;
+   
    SLsmg_gotorc (SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol ();
 }
@@ -85,7 +96,7 @@ int slrn_message (char *fmt, ...)
    va_list ap;
    
    if (Error_Present) return -1;
-   if (Slrn_TT_Initialized)
+   if (Slrn_TT_Initialized & 1)
      {
 	ok = Slrn_Suspension_Ok;
 	if (ok) slrn_set_suspension (0);
@@ -112,7 +123,7 @@ extern void slrn_error (char *fmt, ...)
 {
    va_list ap;
    
-   if (Slrn_TT_Initialized == 0)
+   if ((Slrn_TT_Initialized & 1) == 0)
      {
 	va_start(ap, fmt);
 	(void) vfprintf(stderr, fmt, ap);
@@ -151,7 +162,11 @@ void slrn_print_percent (int row, int col, int n, int nmax, int width)
    SLsmg_printf ("-- %d/%d", n, nmax);
    if (n == 1)
      {
+#if 1
+	SLsmg_write_string ( (n + width > nmax) ? "  (All)" : "  (Top)" );
+#else
 	SLsmg_printf (" (Top)");
+#endif
      }
    else if (n + width > nmax) SLsmg_write_string("  (Bot)");
    else SLsmg_printf("  (%d%%)", ((n + width - 1) * 100) / nmax);
@@ -201,7 +216,15 @@ void slrn_make_home_filename (char *name, char *file)
 {
    char *home;
 #ifndef VMS
-   if ((*name == '/') || ((*name == '.') && (*(name + 1) == '/')))
+   if (
+#ifdef __os2__
+       (*name == '/') || (*name == '\\') || (*(name + 1) == ':') 
+       || ((*name == '.') && (*(name + 1) == '/')) 
+       || ((*name == '.') && (*(name + 1) == '\\'))
+#else
+       (*name == '/') || ((*name == '.') && (*(name + 1) == '/'))
+#endif
+       )
      {
 	strcpy (file, name);
 	return;
@@ -216,55 +239,67 @@ void slrn_make_home_filename (char *name, char *file)
    strcat (file, name);
    
 #else
-   char *cp;
+   char *cp, *cp1;
    static char fname[256];
-   char fn[256];
+   char fn[256], fn1[256];
    int rc, idx;
    
+   strcpy (fn1, name);
    if (NULL != slrn_strchr (name, ':'))
      {
 	strcpy (file, name);
 	return;
      }
-
+   
    if (NULL == (home = getenv ("SLRNHOME")))
      home = getenv ("HOME");
    
    *file = 0;
-   if (cp = strchr(name,'/'))
+   if (NULL != (cp = slrn_strchr (fn1, '/')))
      {
-#ifdef __DECC
-	cp = decc$translate_vms(home);
-	if (cp == 0 || (int)cp == -1)
+# ifdef __DECC
+	*cp = '\0'; cp++;
+	cp1 = decc$translate_vms(home);
+	if (cp1 == 0 || (int)cp1 == -1)
 	  { /* error translating */ }
 	else
 	  {
-	     strcpy(fname, cp);
-	     strcat(cp, "/");
-	  }
-	strcat (cp, name);
-
-	vms_fix_name (cp);
+	     strcpy(fname, cp1);
+	     strcat(cp1, "/");
+ 	  }
+	strcat (cp1, fn1);
 	
-	rc = decc$to_vms(cp, vms_copyname2, 0, 2);
+	vms_fix_name (cp1);
+ 	
+	rc = decc$to_vms(cp1, vms_copyname2, 0, 2);
+ 	if (rc > 0)
+ 	  {
+ 	     strcpy(fname, Copystr);
+ 	     rc = mkdir(fname, 0755);
+ 	  }
+	strcat(fname, cp);
+# else
+	*cp = '\0'; cp++;
+	cp1 = shell$translate_vms(home);
+	if (cp1 == 0 || (int)cp1 == -1)
+	  { /* error translating */ }
+	else
+ 	  {
+	     strcpy(fname, cp1);
+	     strcat(cp1, "/");
+	  }
+	strcat (cp1, fn1);
+	
+	vms_fix_name (cp1);
+ 	
+	rc = shell$to_vms(cp1, vms_copyname2, 0, 2);
 	if (rc > 0)
 	  {
 	     strcpy(fname, Copystr);
 	     rc = mkdir(fname, 0755);
 	  }
-#else
-	if (shell$from_vms(home, vms_copyname1, 0))
-	  {
-	     if (Copystr != NULL) strcpy (fn, Copystr);
-	     strcat(fn, "/");
-	  }
-	
-	strcat (fn, name);
-	vms_fix_name (fn);
-	
-	if (shell$to_vms(fn, vms_copyname1, 0))
-	  strcpy(fname, Copystr);
-#endif
+	strcat(fname, cp);
+# endif
 	strcpy(file,fname);
      }
    else
@@ -361,15 +396,21 @@ FILE *slrn_open_tmpfile (char *file, char *mode)
 	free (f);
      }
 #else
-   char *dir = getenv ("TMPDIR");
+   char *dir;
+
+# ifdef __os2__
+   dir = getenv ("TMP");
+   if (dir == NULL)
+# endif
+     dir = getenv ("TMPDIR");
    
-#ifdef VMS
+# ifdef VMS
    if (dir == NULL) dir = "SYS$LOGIN:";
    sprintf (file, "%sSLRN%uXXXXXX", dir, (unsigned int) getpid ());
-#else
+# else
    if (dir == NULL) dir = "/tmp";
    sprintf (file, "%s/SLRN%uXXXXXX", dir, (unsigned int) getpid ());
-#endif
+# endif
    
    if (NULL != mktemp (file)) fp = fopen (file, mode);
 #endif
@@ -399,13 +440,18 @@ int slrn_posix_system (char *cmd, int reset)
 {
    int init_mode = Slrn_TT_Initialized;
 #ifndef VMS
-   void (*old_int)(int);
+   void (*sint)(int);
+   void (*squit)(int);
    int ret;
    
-   old_int = signal (SIGINT, SIG_IGN);
+   sint = signal (SIGINT, SIG_IGN);
+   squit = signal (SIGQUIT, SIG_IGN);
    if (reset) slrn_reset_display (0);
    ret = system (cmd);
-   signal (SIGINT, old_int);
+   signal (SIGINT, sint);
+# ifdef SIGQUIT
+   signal (SIGQUIT, squit);
+# endif
 #else
    int ret;
    if (reset) slrn_reset_display (0);
@@ -501,6 +547,13 @@ int slrn_edit_file (char *file, unsigned int line)
    if (0 == create_edit_command(editbuf, buf, file, line)) return -1;
    
    ret = slrn_posix_system (buf, 1);
+   
+   /* Am I the only one who thinks this is a good idea?? */
+   if (Slrn_TT_Initialized) while (SLang_input_pending (5))
+     {
+	SLang_flush_input ();
+     }
+   
    /* slrn_redraw (); */
    return ret;
 }
@@ -517,7 +570,7 @@ char slrn_get_response (char *valid_chars, char *str, ...)
    va_list ap;
    char *v;
    
-   if (SLang_Error) return -1;
+   /* if (SLang_Error) return -1; */
    
    while (1)
      {
@@ -543,6 +596,7 @@ char slrn_get_response (char *valid_chars, char *str, ...)
 	     
 	     ch = SLang_getkey ();
 	     slrn_clear_message ();
+	     SLang_Error = SLKeyBoard_Quit = 0;
 	  }
 	
 	v = valid_chars;
@@ -553,7 +607,7 @@ char slrn_get_response (char *valid_chars, char *str, ...)
 	  }
 	
 	slrn_error ("Invalid response! Try again.");
-	if (Slrn_TT_Initialized) 
+	if (Slrn_TT_Initialized & 1)
 	  {
 	     slrn_smg_refresh ();
 	     (void) SLang_input_pending (15);
@@ -568,7 +622,7 @@ int slrn_get_yesno (int dflt, char *str, ...)
    char ch, rsp;
    char *fmt;
    
-   if (SLang_Error) return -1;
+   /* if (SLang_Error) return -1; */
    
    va_start(ap, str);
    (void) vsprintf(buf, str, ap);
@@ -634,20 +688,17 @@ Slrn_User_Info_Type Slrn_User_Info =
      NULL				       /* signature */
 };
 
-#ifndef VMS
-# include <netdb.h>
-  extern int h_errno;
+#if defined(VMS) && defined(MULTINET)
+# include "multinet_root:[multinet.include]netdb.h"
 #else
-# ifdef MULTINET
-#   include "multinet_root:[multinet.include]netdb.h"
-# else
-#   include <netdb.h>
-    extern int h_errno;
+# include <netdb.h>
+# ifndef h_errno
+     extern int h_errno;
 # endif
-#endif /* VMS */
+#endif 
 
 
-static char *make_string (char *s)
+char *slrn_make_startup_string (char *s)
 {
    char *ss;
    ss = (char *) SLMALLOC (strlen (s) + 1);
@@ -713,7 +764,8 @@ int slrn_is_fqdn (char *h)
     * id will be composed from the fqdn.  For that reason, such names will
     * be rejected.  Sigh.
     */
-   if (NULL != slrn_strchr (h, '('))
+   if ((NULL != slrn_strchr (h, '('))
+       || (NULL != slrn_strchr (h, '@')))
      return 0;
    
    p = slrn_strchr (h, '.');
@@ -725,15 +777,15 @@ void slrn_get_user_info (void)
 {
    struct hostent *host_entry;
    char *name, *host;
-#ifdef unix
+#ifndef VMS
    struct passwd *pw;
    
    pw = getpwuid (getuid ());
 #endif
    
    /* Fill in what is assumed to be non-NULL by rest of program. */
-   Slrn_User_Info.followup_string = make_string ("In article %m, %r wrote:");
-   Slrn_Courtesy_CC_Message = make_string ("[This message has also been posted.]");
+   Slrn_User_Info.followup_string = slrn_make_startup_string ("In article %m, %r wrote:");
+   Slrn_Courtesy_CC_Message = slrn_make_startup_string ("[This message has also been posted.]");
    
    /* Now get default values for rest. */
    
@@ -760,7 +812,7 @@ void slrn_get_user_info (void)
      {
 	if (slrn_is_fqdn ((char *)host_entry->h_name))
 	  {
-	     Slrn_User_Info.posting_host = make_string ((char *)host_entry->h_name);
+	     Slrn_User_Info.posting_host = slrn_make_startup_string ((char *)host_entry->h_name);
 	  }
 	else
 	  {
@@ -770,7 +822,7 @@ void slrn_get_user_info (void)
 	       {
 		  if (slrn_is_fqdn (*aliases))
 		    {
-		       Slrn_User_Info.posting_host = make_string (*aliases);
+		       Slrn_User_Info.posting_host = slrn_make_startup_string (*aliases);
 		       break;
 		    }
 		  aliases++;
@@ -784,14 +836,56 @@ void slrn_get_user_info (void)
 	else strcpy (host, Slrn_User_Info.posting_host);
      }
    
-   
+#if defined(USE_DOMAIN_NAME) && defined(MY_DOMAIN_NAME)
+   if (*host && (0 == slrn_is_fqdn (host)))
+     {
+	char *my_domain_name;
+	unsigned int len;
+	
+	len = strlen (host);
+	my_domain_name = MY_DOMAIN_NAME;
+	if (*my_domain_name != '.')
+	  {
+	     host[len++] = '.';
+	  }
+	strcpy (host + len, my_domain_name);
+     }
+#endif
    /* If a value was compiled in, use it. */
 #ifdef OUR_HOSTNAME
-   if (slrn_is_fqdn (OUR_HOSTNAME))
-     strcpy (host, OUR_HOSTNAME);
+# ifdef unix
+   /* If the value is a file, read the file for the FQDN */
+   if (OUR_HOSTNAME[0] == '/')
+     {
+	FILE *fp;
+	
+	if (NULL != (fp = fopen(OUR_HOSTNAME, "r")))
+	  {
+	   char tmphost[MAX_HOST_NAME_LEN];
+	     
+	     if (fgets(tmphost, MAX_HOST_NAME_LEN , fp) != NULL)
+	       {
+		  char *p;
+		  
+		  if ((p = slrn_strchr (tmphost, '\n')) != NULL)
+		    *p = '\0';
+	       /* If the entry is valid, we're done */
+		  if (slrn_is_fqdn(tmphost))
+		    strcpy(host, tmphost);
+	       }
+	     fclose(fp);
+	  }
+     }
+   else
+# endif
+     if (slrn_is_fqdn (OUR_HOSTNAME)
+	 && strlen(OUR_HOSTNAME) < MAX_HOST_NAME_LEN)
+       strcpy (host, OUR_HOSTNAME);
 #endif
-   
-   /* Allow user chance to over ride this */
+
+   /* Allow user chance to over ride this.  Until now, host was a pointer
+    * to Slrn_User_Info.host.
+    */
    if ((NULL != (host = getenv ("HOSTNAME")))
        && slrn_is_fqdn (host))
      {
@@ -816,11 +910,11 @@ void slrn_get_user_info (void)
      name = "";
 
 
-   Slrn_User_Info.username = make_string (name);
+   Slrn_User_Info.username = slrn_make_startup_string (name);
    
    if ((Slrn_User_Info.replyto = getenv ("REPLYTO")) == NULL)
      Slrn_User_Info.replyto = "";
-   Slrn_User_Info.replyto = make_string (Slrn_User_Info.replyto);
+   Slrn_User_Info.replyto = slrn_make_startup_string (Slrn_User_Info.replyto);
    
 #ifdef VMS
    Slrn_User_Info.realname = fix_fullname(get_uaf_fullname());
@@ -831,7 +925,7 @@ void slrn_get_user_info (void)
 	Slrn_User_Info.realname = "";
      }
 #endif
-   Slrn_User_Info.realname = make_string (Slrn_User_Info.realname);
+   Slrn_User_Info.realname = slrn_make_startup_string (Slrn_User_Info.realname);
    
    /* truncate at character used to delineate extra gecos fields */
    name = Slrn_User_Info.realname;
@@ -861,17 +955,17 @@ void slrn_get_user_info (void)
 		  slrn_fclose (fporg);
 	       }
 	  }
-	Slrn_User_Info.org = make_string (Slrn_User_Info.org);
+	Slrn_User_Info.org = slrn_make_startup_string (Slrn_User_Info.org);
      }
    
-   Slrn_User_Info.signature = make_string (SLRN_SIGNATURE_FILE);
+   Slrn_User_Info.signature = slrn_make_startup_string (SLRN_SIGNATURE_FILE);
    
-#ifdef HAS_MIME_SUPPORT
-   Slrn_Mime_Display_Charset = make_string ("iso-8859-1");
+#if SLRN_HAS_MIME
+   Slrn_Mime_Display_Charset = slrn_make_startup_string ("iso-8859-1");
 #endif
    
 #ifdef SLRN_SENDMAIL_COMMAND
-   Slrn_SendMail_Command = make_string (SLRN_SENDMAIL_COMMAND);
+   Slrn_SendMail_Command = slrn_make_startup_string (SLRN_SENDMAIL_COMMAND);
 #endif
 }
 
@@ -884,7 +978,7 @@ void update_top_status_line (void)
    SLsmg_printf ("\
 slrn %s ** Press '?' for help, 'q' to quit. ** Server: %s",
 		 Slrn_Version,
-		 Slrn_Server_Name);
+		 Slrn_Server_Obj->name);
    SLsmg_erase_eol ();
    slrn_set_color (0);
 }
@@ -908,8 +1002,9 @@ static SLang_RLine_Info_Type  *init_readline (void)
    if ((NULL == (rli = (SLang_RLine_Info_Type *) SLMALLOC (sizeof(SLang_RLine_Info_Type))))
        || (NULL == (buf = (unsigned char *) SLMALLOC (256))))
      {
+	if (rli != NULL) SLFREE (rli);
 	fprintf(stderr, "malloc error.\n");
-	exit(-1);
+	return NULL;
      }
    
    MEMSET ((char *) rli, 0, sizeof (SLang_RLine_Info_Type));
@@ -923,22 +1018,26 @@ static SLang_RLine_Info_Type  *init_readline (void)
    
    if (SLang_init_readline (rli) < 0)
      {
-	fprintf(stderr, "Unable to initialize readline library.\n");
-	exit (-1);
+	SLFREE (rli);
+	SLFREE (buf);
+	rli = NULL;
      }
    
    return rli;
 }
 
+int slrn_init_readline (void)
+{
+   if ((Slrn_Keymap_RLI == NULL) 
+       && (NULL == (Slrn_Keymap_RLI = init_readline ())))
+     return -1;
+   return 0;
+}   
+
 int slrn_read_input (char *prompt, char *str, int trim_flag)
 {
    int i;
    
-   if (Slrn_Keymap_RLI == NULL)
-     {
-	Slrn_Keymap_RLI = init_readline ();
-     }
-   
    Slrn_Keymap_RLI->edit_width = SLtt_Screen_Cols - 1;
    Slrn_Keymap_RLI->prompt = prompt;
    *Slrn_Keymap_RLI->buf = 0;
@@ -955,11 +1054,14 @@ int slrn_read_input (char *prompt, char *str, int trim_flag)
    
    if ((i >= 0) && !SLang_Error && !SLKeyBoard_Quit)
      {
-	char *b;
+	char *b = (char *) Slrn_Keymap_RLI->buf;
 	
 	/* SLang_rline_save_line (Slrn_Keymap_RLI); */
-	slrn_trim_string ((char *) Slrn_Keymap_RLI->buf);
-	b = slrn_skip_whitespace ((char *) Slrn_Keymap_RLI->buf);
+	if (trim_flag) 
+	  {
+	     slrn_trim_string (b);
+	     b = slrn_skip_whitespace (b);
+	  }
 	strcpy (str, b);
 	i = strlen (str);
      }
@@ -984,16 +1086,12 @@ int slrn_read_integer (char *prompt, int *dflt, int *np)
      }
    else strcpy (buf, prompt);
    
-   do
+   *str = 0;
+   if (-1 == (n = slrn_read_input (buf, str, 1)))
      {
-	*str = 0;
-	if (-1 == (n = slrn_read_input (buf, str, 1)))
-	  {
-	     slrn_error ("Abort!");
-	     return -1;
-	  }
+	slrn_error ("Abort!");
+	return -1;
      }
-   while ((n == 0) && (dflt == NULL));
    
    if (1 != sscanf(str, "%d", &n))
      {
@@ -1114,55 +1212,29 @@ char *slrn_fix_regexp (char *pat)
 }
 
 
-/* Note!!!  These routines assume a flat address space !! */
-int slrn_case_strncmp (unsigned char *a, register unsigned char *b, register unsigned int n)
-{
-   register unsigned char cha, chb, *bmax;
-   register int diff = a - b;
-   
-   bmax = b + n;
-   while (b < bmax)
-     {
-	cha = UPPER_CASE(b[diff]);
-	chb = UPPER_CASE(*b);
-	if (cha != chb)
-	  {
-	     return (int) cha - (int) chb;
-	  }
-	else if (chb == 0) return 0;
-	b++;
-     }
-   return 0;
-}
-
-int slrn_case_strcmp (unsigned char *a, register unsigned char *b)
-{
-   register unsigned char cha, chb;
-   register int diff = a - b;
-   
-   while (1)
-     {
-	cha = UPPER_CASE(b[diff]);
-	chb = UPPER_CASE(*b);
-	if (cha != chb)
-	  {
-	     return (int) cha - (int) chb;
-	  }
-	else if (chb == 0) break;
-	b++;
-     }
-   return 0;
-}
-
-
 int slrn_file_exists (char *file)
 {
    struct stat st;
    int m;
    
+#ifdef _S_IFDIR
+# ifndef S_IFDIR
+#  define S_IFDIR _S_IFDIR
+# endif
+#endif
+   
+#ifndef S_ISDIR
+# ifdef S_IFDIR
+#  define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
+# else
+#  define S_ISDIR(m) 0
+# endif
+#endif
+   
    if (stat(file, &st) < 0) return 0;
    m = st.st_mode;
-   if (m & S_IFDIR) return (2);
+   
+   if (S_ISDIR(m)) return (2);
    return 1;
 }
 
@@ -1177,7 +1249,7 @@ void slrn_get_mouse_rc (int *rp, int *cp)
    if (rp != NULL) *rp = r;
 }
 
-#ifndef NO_PIPING
+#if SLRN_HAS_PIPING
 int slrn_pclose (FILE *fp)
 {
    int ret;
