@@ -1,7 +1,8 @@
 /*  Copyright (c) 1995 John E. Davis (davis@space.mit.edu)
  *  All rights reserved.
  */
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -10,57 +11,96 @@
 
 
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+# include <unistd.h>
 #endif
 
 #include <slang.h>
 
-#include "clientlib.h"
-#include "nntp.h"
-#include "misc.h"
 #include "slrn.h"
+#include "server.h"
+#include "misc.h"
 
-int Slrn_Can_Post;
+static Slrn_Server_Obj_Type NNTP_Server_Obj;
+static Slrn_Post_Obj_Type NNTP_Post_Obj;
 
-int Slrn_NNTP_Has_XOver;
-static int Has_XPat = -1;	       /* -1 means probe needs performed */
-static int Has_Xhdr = -1;
-static int Has_Xgtitle = -1;
+char *Slrn_NNTP_Server_Name;
 
-int slrn_put_server (char *line)
+int Slrn_Query_Reconnect = 1;
+int Slrn_NNTP_Port = -1;
+
+static int nntp_select_group (char *, int *, int *);
+static char *nntp_read_line (char *, unsigned int);
+static void nntp_close_server (void);
+static int nntp_initialize_server (void);
+static int nntp_select_article (int, char *);
+static char *nntp_head_from_msgid (char *, char *, unsigned int);
+static char *nntp_read_xover (char *, unsigned int);
+static int nntp_open_xover (int, int);
+static void nntp_close_xover (void);
+static int nntp_xpat_cmd (char *, int, int, char *);
+static int nntp_printf (char *, ...);
+static int nntp_puts (char *);
+static int nntp_xhdr_command (char *, int, char *, unsigned int);
+static char *nntp_get_extra_xover_header (char *);
+static void nntp_close_suspend_xover (void);
+static void nntp_open_suspend_xover (void);
+static int nntp_xgtitle_cmd (char *);
+static int nntp_has_cmd (char *);
+static int nntp_list_newsgroups (void);
+static int nntp_list_active (void);
+
+
+static int NNTP_Has_XPat = -1;	       /* -1 means probe needs performed */
+static int NNTP_Has_Xhdr = -1;
+static int NNTP_Has_Xgtitle = -1;
+
+static char *nntp_read_line (char *, unsigned int);
+
+static int NNTP_Reconnect_Ok;
+static char NNTP_Last_Reconnect_Command[256];
+
+static int slrn_get_server (char *, unsigned int, int);
+
+/* This function does not attempt a reconnection.  A reconnection can be
+ * achieved only in a definite context.  
+ */
+static int slrn_put_server_cmd (char *line, char *buf, unsigned int len)
 {
-   if (-1 ==client_put_server (line))
+   NNTP_Reconnect_Ok = 0;
+   if (-1 == client_put_server (line))
      {
 	slrn_exit_error ("Server closed connection on %s command.", line);
      }
-   return 0;
+   *buf = 0;
+   slrn_get_server (buf, len, 1);
+   return atoi (buf);
 }
 
 
-int slrn_get_server (char *line, int len, int fatal)
+static int slrn_get_server (char *line, unsigned int len, int fatal)
 {
    int ret;
    *line = 0;
    while (1)
      {
 	errno = 0;
-	ret = client_get_server (line, len);
+	ret = client_get_server (line, (int) len);
 	if (ret == -1)
 	  {
-#ifdef EAGAIN
+# ifdef EAGAIN
 	     if (errno == EAGAIN)
 	       {
 		  sleep (1);
 		  continue;
 	       }
-#endif
-#ifdef EINTR
+# endif
+# ifdef EINTR
 	     if (errno == EINTR)
 	       {
 		  slrn_error ("read aborted.");
 		  return -1;
 	       }
-#endif
+# endif
 	     if (fatal) slrn_exit_error ("Read failed from server. (%d)", errno);
 	     return -1;
 	  }
@@ -71,14 +111,21 @@ int slrn_get_server (char *line, int len, int fatal)
 static int nntp_attempt_reconnect (void)
 {
    char buf[256];
+   int max_tries_without_query = 1;
    
    while (1)
      {
-	if (slrn_get_yesno (1, "NNTP connection dropped -- reconnect") <= 0)
+	if ((Slrn_Query_Reconnect || (max_tries_without_query == 0))
+	    && (slrn_get_yesno (1, "NNTP connection dropped -- reconnect") <= 0))
 	  return -1;
 	
-	if (nntp_initialize_server (Slrn_Server_Name) != 0)
-	  continue;
+	if (max_tries_without_query != 0) max_tries_without_query--;
+	
+	if (nntp_initialize_server () != 0)
+	  {
+	     continue;
+	  }
+	
 	
 	if ((NULL != Slrn_Current_Group) && ('\0' != *Slrn_Current_Group))
 	  {
@@ -95,16 +142,22 @@ static int nntp_attempt_reconnect (void)
 
 static int put_server_with_reconnect (char *cmd, char *buf, unsigned int len)
 {
+   NNTP_Reconnect_Ok = 1;
+   strcpy (NNTP_Last_Reconnect_Command, cmd);
+   
    while (1)
      {
 	*buf = 0;
-	if ((-1 == client_put_server (cmd))
+	if ((-1 == client_put_server (NNTP_Last_Reconnect_Command))
 	    || (-1 == slrn_get_server (buf, len - 1, 0))
-	    || (ERR_FAULT == atoi (buf)))
+# if 0
+	    || (ERR_FAULT == atoi (buf))
+# endif
+	    )
 	  {
 	     if (-1 == nntp_attempt_reconnect ())
 	       {
-		  slrn_exit_error ("Server connection closed. (%s)", cmd);
+		  slrn_exit_error ("Server connection closed. (%s)", NNTP_Last_Reconnect_Command);
 	       }
 	  }
 	else break;
@@ -113,19 +166,27 @@ static int put_server_with_reconnect (char *cmd, char *buf, unsigned int len)
    return atoi(buf);
 }
 
+static void nntp_discard_output (char *line, unsigned int len)
+{
+   while (NULL != nntp_read_line (line, len))
+     {
+	;
+     }
+}
+
 
 int nntp_select_group (char *name, int *min, int *max)
 {
-   char line[256], group[256];
+   char group[256];
    int estim;
    int code;
    
    sprintf (group, "GROUP %s", name);
    
-   code = put_server_with_reconnect (group, line, sizeof (line));
+   code = put_server_with_reconnect (group, group, sizeof (group));
    
    /* responses 211 and 411 as well as 502 */
-   if (code != OK_GROUP) 
+   if (code != OK_GROUP)
      {
 	if (code == ERR_ACCESS)
 	  {
@@ -135,16 +196,17 @@ int nntp_select_group (char *name, int *min, int *max)
 	return -1;
      }
    
-   if (3 != sscanf(line + 4, "%d %d %d", &estim, min, max)) return -1;
-
+   if (3 != sscanf(group + 4, "%d %d %d", &estim, min, max)) return -1;
+   
    return 0;
 }
 
-static int Server_Inited = 0;
+static int NNTP_Server_Inited = 0;
+
 void nntp_close_server (void)
 {
    client_close_server ();
-   Server_Inited = 0;
+   NNTP_Server_Inited = 0;
 }
 
 static int nntp_authorization (void)
@@ -152,20 +214,17 @@ static int nntp_authorization (void)
    char line[NNTP_STRLEN];
    
    sprintf (line, "authinfo user %s", Slrn_User_Info.nnrpname);
-   slrn_put_server (line);
-   slrn_get_server(line, sizeof(line), 1);
    
-   if (atoi(line) == NEED_AUTHDATA)
+   if (NEED_AUTHDATA == slrn_put_server_cmd (line, line, sizeof (line)))
      {
 	sprintf (line, "authinfo pass %s", Slrn_User_Info.nnrppass);
-	slrn_put_server (line);
-	slrn_get_server(line, sizeof(line), 1);
-	switch (atoi(line))
+	
+	switch (slrn_put_server_cmd (line, line, sizeof (line)))
 	  {
 	   case ERR_ACCESS:
 	     return -1;
 	   case OK_AUTH:
-	     Slrn_Can_Post = 1;
+	     Slrn_Post_Obj->can_post = 1;
 	     return 0;
 	  }
      }
@@ -181,24 +240,24 @@ static int probe_server (char *cmd)
    return 1;
 }
 
-#define PROBE_XCMD(var, cmd) (((var) != -1) ? (var) : ((var) = probe_server(cmd)))
+# define PROBE_XCMD(var, cmd) (((var) != -1) ? (var) : ((var) = probe_server(cmd)))
 
 int nntp_has_cmd (char *cmd)
 {
    if (!strcmp (cmd, "XHDR"))
-     return PROBE_XCMD(Has_Xhdr, cmd);
+     return PROBE_XCMD(NNTP_Has_Xhdr, cmd);
    
    if (!strcmp (cmd, "XPAT"))
-     return PROBE_XCMD(Has_XPat, cmd);
+     return PROBE_XCMD(NNTP_Has_XPat, cmd);
    
    if (!strcmp (cmd, "XGTITLE"))
-     return PROBE_XCMD(Has_Xgtitle, cmd);
-       
+     return PROBE_XCMD(NNTP_Has_Xgtitle, cmd);
+   
    return probe_server (cmd);
 }
 
 
-int nntp_initialize_server (char *host)
+static int nntp_initialize_server (void)
 {
    char line[256];
    int code;
@@ -207,8 +266,9 @@ int nntp_initialize_server (char *host)
    char msg2[] = "Unable to initialize server: %s";
    char msg4[] = "Authenticating %s...";
    int tt_inited = (Slrn_TT_Initialized & 1);
+   char *host = Slrn_NNTP_Server_Name;
    
-   if (Server_Inited) nntp_close_server ();
+   if (NNTP_Server_Inited) nntp_close_server ();
    
    if (tt_inited == 0)
      slrn_tty_message (0, msg1, host);
@@ -229,16 +289,16 @@ int nntp_initialize_server (char *host)
 	  }
 	return -1;
      }
-   Server_Inited = 1;
+   NNTP_Server_Inited = 1;
    
    switch (code)
      {
       case OK_CANPOST:
-	Slrn_Can_Post = 1;
+	Slrn_Post_Obj->can_post = 1;
 	str = "Posting ok.";
 	break;
       case OK_NOPOST:
-	Slrn_Can_Post = 0;
+	Slrn_Post_Obj->can_post = 0;
 	str = "Posting NOT ok.";
 	break;
       default:
@@ -263,10 +323,8 @@ int nntp_initialize_server (char *host)
     * process that actually performs the role of server.  It is what needs
     * authorization.
     */
-   slrn_put_server ("mode reader");
-   slrn_get_server (line, sizeof(line), 1);
    
-   if (atoi(line) == ERR_ACCESS)
+   if (ERR_ACCESS == slrn_put_server_cmd ("mode reader", line, sizeof (line)))
      {
 	char msg3[] = "NNTP: 'mode reader' failed: %s";
 	if (tt_inited)
@@ -277,7 +335,7 @@ int nntp_initialize_server (char *host)
 	else slrn_tty_message (2, msg3);
  	return (-1);
      }
-
+   
    if ((Slrn_User_Info.nnrpname != NULL)
        && (*Slrn_User_Info.nnrpname != 0))
      {
@@ -327,14 +385,11 @@ int nntp_initialize_server (char *host)
 	     else slrn_tty_message (0, "Succeeded");
 	  }
      }
-
-   slrn_put_server ("XOVER");
-   slrn_get_server (line, sizeof(line), 1);
    
-   if (atoi(line) == ERR_COMMAND)
+   if (ERR_COMMAND == slrn_put_server_cmd ("XOVER", line, sizeof (line)))
      {
 	char msg5[] = "Server %s does not implement the XOVER command.";
-	Slrn_NNTP_Has_XOver = 0;
+	NNTP_Server_Obj.has_xover = 0;
 	if (tt_inited)
 	  {
 	     slrn_message (msg5, host);
@@ -342,103 +397,101 @@ int nntp_initialize_server (char *host)
 	  }
 	else slrn_tty_message (2, msg5, host);
      }
-   else Slrn_NNTP_Has_XOver = 1;
-   
+   else NNTP_Server_Obj.has_xover = 1;
    return 0;
 }
 
-char *nntp_read_line (char *line, int len)
+static char *nntp_read_line (char *line, unsigned int len)
 {
+   /* This needs to be modified if Reconnection_Ok is non-zero.  Somehow,
+    * the calling routine needs to be told to re-issue the NNTP command 
+    * and restart.
+    */
    slrn_get_server (line, len, 1);
    
    if ((*line == '.') && (*(line + 1) == 0)) return NULL;
    return line;
 }
 
-int nntp_xpat_cmd (char *hdr, int rmin, int rmax, char *pat)
+static int nntp_xpat_cmd (char *hdr, int rmin, int rmax, char *pat)
 {
    char buf[512];
    
-   if (0 == PROBE_XCMD(Has_XPat, "XPAT"))
+   if (0 == PROBE_XCMD(NNTP_Has_XPat, "XPAT"))
      return -1;
    
    sprintf (buf, "XPAT %s %d-%d *%s*", hdr, rmin, rmax, pat);
-   slrn_put_server (buf);
-   if (-1 == slrn_get_server (buf, sizeof (buf) - 1, 0))
+   if (OK_HEAD != put_server_with_reconnect (buf, buf, sizeof (buf)))
      return -1;
-   if (atoi (buf) != OK_HEAD) return -1;
    return 0;
 }
 
-int nntp_xgtitle_cmd (char *pattern)
+static int nntp_xgtitle_cmd (char *pattern)
 {
    char buf[512];
    
    /* XGTITLE appears broken on some servers.  So, do not probe for it. */
-   if (Has_Xgtitle == 0) return -1;
+   if (NNTP_Has_Xgtitle == 0) return -1;
    
-   if (Has_Xgtitle == -1)
+   if (NNTP_Has_Xgtitle == -1)
      {
 	slrn_message ("Sending Query to Server...");
 	slrn_smg_refresh ();
+	Slrn_Message_Present = 0;
      }
    
    sprintf (buf, "XGTITLE %s", pattern);
-   slrn_put_server (buf);
-   if (-1 == slrn_get_server (buf, sizeof (buf), 0))
-     return -1;
    
-   if (OK_XGTITLE != atoi (buf))
+   if (OK_XGTITLE != put_server_with_reconnect (buf, buf, sizeof (buf)))
      {
 	slrn_message ("Server does not support XGTITLE command.");
-	Has_Xgtitle = 0;
+	NNTP_Has_Xgtitle = 0;
 	return -1;
      }
    
-   Has_Xgtitle = 1;
+   NNTP_Has_Xgtitle = 1;
    return 0;
 }
 
-int nntp_select_article (int n, char *msgid)
+static int nntp_select_article (int n, char *msgid)
 {
-   char buf[256], cmd[256];
+   char cmd[256];
    
    if (n != -1)
      sprintf (cmd, "ARTICLE %d", n);
    else
      sprintf (cmd, "ARTICLE %s", msgid);
    
-   if (OK_ARTICLE == put_server_with_reconnect (cmd, buf, sizeof (buf)))
+   if (OK_ARTICLE == put_server_with_reconnect (cmd, cmd, sizeof (cmd)))
      return 0;
    else return -1;
 }
 
-static int XOver_Done;
-static int XOver_Min;
-static int XOver_Max;
-static int XOver_Next;
+static int NNTP_XOver_Done;
+static int NNTP_XOver_Min;
+static int NNTP_XOver_Max;
+static int NNTP_XOver_Next;
 
-static int Suspend_XOver_For_Kill = 0;
+static int NNTP_Suspend_XOver_For_Kill = 0;
 
-int nntp_open_xover (int min, int max)
+/* We cannot reconnect on this because of the context may be too complex. */
+static int nntp_open_xover (int min, int max)
 {
    char buf[512];
-   if (Slrn_NNTP_Has_XOver && !Suspend_XOver_For_Kill)
+   if (NNTP_Server_Obj.has_xover && !NNTP_Suspend_XOver_For_Kill)
      {
 	sprintf (buf, "XOVER %d-%d", min, max);
-	slrn_put_server (buf);
-	*buf = 0;
-	slrn_get_server (buf, 512, 1);
-	if (atoi(buf) != OK_XOVER)
+	
+	if (OK_XOVER != slrn_put_server_cmd (buf, buf, sizeof (buf)))
 	  {
-	     XOver_Done = 1;
+	     NNTP_XOver_Done = 1;
 	     slrn_error ("XOVER failed.");
 	     return -1;
 	  }
      }
-   XOver_Next = XOver_Min = min;
-   XOver_Max = max;
-   XOver_Done = 0;
+   NNTP_XOver_Next = NNTP_XOver_Min = min;
+   NNTP_XOver_Max = max;
+   NNTP_XOver_Done = 0;
    return 0;
 }
 
@@ -446,14 +499,14 @@ int nntp_open_xover (int min, int max)
  * Relay-Version,  Posting-Version,  From,  Date, Newsgroups,
  * Subject,  Message-ID,  Path. */
 
-#define MAX_XOVER_SIZE 1024
+#  define NNTP_MAX_XOVER_SIZE 1024
 typedef struct
 {
    char name[25];
    unsigned int colon;
-   char buf[MAX_XOVER_SIZE + 16];
+   char buf[NNTP_MAX_XOVER_SIZE + 16];
 }
-Header_Xover_Type;
+NNTP_Header_Xover_Type;
 
 /* These MUST be arranged in XOver format:
  * id|subj|from|date|msgid|refs|bytes|line|misc stuff
@@ -461,8 +514,8 @@ Header_Xover_Type;
  * !!!DO NOT REARRANGE OR ADD ANYTHING WITHOUT UPDATING THE ARRAY REFERENCES!!!
  */
 
-#define NUM_XOVER_HEADERS 7
-Header_Xover_Type Xover_Headers[NUM_XOVER_HEADERS] =
+#  define NUM_XOVER_HEADERS 7
+static NNTP_Header_Xover_Type NNTP_Xover_Headers[NUM_XOVER_HEADERS] =
 {
      {"Subject:", 8, ""},
      {"From:", 5, ""},
@@ -475,38 +528,38 @@ Header_Xover_Type Xover_Headers[NUM_XOVER_HEADERS] =
 };
 
 /* The rest of these are for scoring/killing */
-#define MAX_EXTRA_XOVER_HEADERS 20
-Header_Xover_Type *Extra_Xover_Headers[MAX_EXTRA_XOVER_HEADERS];
-static unsigned int Next_Extra_Xover_Header;
+#  define NNTP_MAX_EXTRA_XOVER_HEADERS 20
+static NNTP_Header_Xover_Type *NNTP_Extra_Xover_Headers[NNTP_MAX_EXTRA_XOVER_HEADERS];
+static unsigned int NNTP_Next_Extra_Xover_Header;
 
-void nntp_open_suspend_xover (void)
+static void nntp_open_suspend_xover (void)
 {
-   Suspend_XOver_For_Kill = 1;
-   Next_Extra_Xover_Header = 0;
+   NNTP_Suspend_XOver_For_Kill = 1;
+   NNTP_Next_Extra_Xover_Header = 0;
 }
 
-void nntp_close_suspend_xover (void)
+static void nntp_close_suspend_xover (void)
 {
    unsigned int i;
    
-   if (Suspend_XOver_For_Kill == 0) return;
-   Suspend_XOver_For_Kill = 0;
-   for (i = 0; i < MAX_EXTRA_XOVER_HEADERS; i++)
+   if (NNTP_Suspend_XOver_For_Kill == 0) return;
+   NNTP_Suspend_XOver_For_Kill = 0;
+   for (i = 0; i < NNTP_MAX_EXTRA_XOVER_HEADERS; i++)
      {
-	if (Extra_Xover_Headers[i] != NULL)
-	  SLFREE (Extra_Xover_Headers[i]);
-	Extra_Xover_Headers[i] = NULL;
+	if (NNTP_Extra_Xover_Headers[i] != NULL)
+	  SLFREE (NNTP_Extra_Xover_Headers[i]);
+	NNTP_Extra_Xover_Headers[i] = NULL;
      }
-   Next_Extra_Xover_Header = 0;
+   NNTP_Next_Extra_Xover_Header = 0;
 }
 
-char *nntp_get_extra_xover_header (char *hdr)
+static char *nntp_get_extra_xover_header (char *hdr)
 {
    unsigned int i;
    unsigned int colon = strlen (hdr) + 1;
-   for (i = 0; i < Next_Extra_Xover_Header; i++)
+   for (i = 0; i < NNTP_Next_Extra_Xover_Header; i++)
      {
-	Header_Xover_Type *xov = Extra_Xover_Headers[i];
+	NNTP_Header_Xover_Type *xov = NNTP_Extra_Xover_Headers[i];
 	if ((colon == xov->colon)
 	    && (!slrn_case_strcmp ((unsigned char *)xov->name,
 				   (unsigned char *)hdr)))
@@ -520,18 +573,18 @@ char *nntp_get_extra_xover_header (char *hdr)
 static char *read_create_xover_line (int id, char *the_buf)
 {
    char *hbuf, *b;
-   char buf[8 * MAX_XOVER_SIZE];
+   char buf[8 * NNTP_MAX_XOVER_SIZE];
    unsigned int colon = 0;
    char *p, ch;
    unsigned int group_len;
    int i, headers_not_found;
-   Header_Xover_Type *xov = NULL;
+   NNTP_Header_Xover_Type *xov = NULL;
    
    for (i = 0; i < NUM_XOVER_HEADERS; i++)
      {
-	*Xover_Headers[i].buf = 0;
+	*NNTP_Xover_Headers[i].buf = 0;
      }
-   Next_Extra_Xover_Header = 0;
+   NNTP_Next_Extra_Xover_Header = 0;
    
    headers_not_found = NUM_XOVER_HEADERS;
    
@@ -545,7 +598,7 @@ static char *read_create_xover_line (int id, char *the_buf)
 	if ((ch == ' ') || (ch == '\t'))
 	  {
 	     if (hbuf == NULL) continue;
-	     if (hbuf - xov->buf >= MAX_XOVER_SIZE - 2)
+	     if (hbuf - xov->buf >= NNTP_MAX_XOVER_SIZE - 2)
 	       {
 		  hbuf = NULL;
 		  continue;
@@ -563,13 +616,13 @@ static char *read_create_xover_line (int id, char *the_buf)
 	     
 	     while (i < NUM_XOVER_HEADERS)
 	       {
-		  if ((ch == *Xover_Headers[i].name)
-		      && !slrn_case_strncmp ((unsigned char *)Xover_Headers[i].name,
-					     (unsigned char *)buf, Xover_Headers[i].colon))
+		  if ((ch == *NNTP_Xover_Headers[i].name)
+		      && !slrn_case_strncmp ((unsigned char *)NNTP_Xover_Headers[i].name,
+					     (unsigned char *)buf, NNTP_Xover_Headers[i].colon))
 		    {
 		       headers_not_found--;
 		       
-		       xov = Xover_Headers + i;
+		       xov = NNTP_Xover_Headers + i;
 		       colon = xov->colon;
 		       hbuf = xov->buf;
 		       break;
@@ -579,23 +632,23 @@ static char *read_create_xover_line (int id, char *the_buf)
 	     
 	     if (i == NUM_XOVER_HEADERS)
 	       {
-		  if ((!Suspend_XOver_For_Kill)
-		      || (Next_Extra_Xover_Header == MAX_EXTRA_XOVER_HEADERS))
+		  if ((!NNTP_Suspend_XOver_For_Kill)
+		      || (NNTP_Next_Extra_Xover_Header == NNTP_MAX_EXTRA_XOVER_HEADERS))
 		    {
 		       hbuf = NULL;
 		       continue;
 		    }
 		  
-		  xov = Extra_Xover_Headers[Next_Extra_Xover_Header];
+		  xov = NNTP_Extra_Xover_Headers[NNTP_Next_Extra_Xover_Header];
 		  if (xov == NULL)
 		    {
-		       xov = (Header_Xover_Type *) SLMALLOC (sizeof(Header_Xover_Type));
+		       xov = (NNTP_Header_Xover_Type *) SLMALLOC (sizeof(NNTP_Header_Xover_Type));
 		       if (xov == NULL)
 			 {
 			    hbuf = NULL;
 			    continue;
 			 }
-		       Extra_Xover_Headers[Next_Extra_Xover_Header] = xov;
+		       NNTP_Extra_Xover_Headers[NNTP_Next_Extra_Xover_Header] = xov;
 		    }
 		  
 		  b = buf;
@@ -612,7 +665,7 @@ static char *read_create_xover_line (int id, char *the_buf)
 		  xov->name[sizeof(xov->name) - 1] = 0;
 		  xov->colon = colon;
 		  
-		  Next_Extra_Xover_Header += 1;
+		  NNTP_Next_Extra_Xover_Header += 1;
 		  hbuf = xov->buf;
 	       }
 	  }
@@ -625,7 +678,7 @@ static char *read_create_xover_line (int id, char *the_buf)
 	/* trim leading whitespace */
 	b = slrn_skip_whitespace (buf + colon);
 	
-	maxlen = (MAX_XOVER_SIZE - (unsigned int)(hbuf - xov->buf)) - 1;
+	maxlen = (NNTP_MAX_XOVER_SIZE - (unsigned int)(hbuf - xov->buf)) - 1;
 	strncpy (hbuf, b, maxlen);
 	hbuf[maxlen] = 0;
 	
@@ -639,7 +692,7 @@ static char *read_create_xover_line (int id, char *the_buf)
    /* we should try to extract the id from the Xref header if possible */
    if (id == -1)
      {
-	char *xref = Xover_Headers[6].buf;
+	char *xref = NNTP_Xover_Headers[6].buf;
 	if (*xref != 0)
 	  {
 	     group_len = strlen (Slrn_Current_Group);
@@ -666,10 +719,10 @@ static char *read_create_xover_line (int id, char *the_buf)
 	       }
 	  }
 	
-	if ((id == -1) && PROBE_XCMD(Has_XPat, "XPAT"))
+	if ((id == -1) && PROBE_XCMD(NNTP_Has_XPat, "XPAT"))
 	  {
 	     char xpatbuf[256];
-	     char *msgid = Xover_Headers[3].buf;
+	     char *msgid = NNTP_Xover_Headers[3].buf;
 	     if (-1 != nntp_xpat_cmd ("Message-Id", Slrn_Server_Min, Slrn_Server_Max, msgid))
 	       {
 		  /* This should only loop once. */
@@ -688,37 +741,35 @@ static char *read_create_xover_line (int id, char *the_buf)
    
    sprintf (the_buf, "%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tXref: %s",
 	    id,
-	    Xover_Headers[0].buf,      /* subject */
-	    Xover_Headers[1].buf,      /* from */
-	    Xover_Headers[2].buf,      /* date */
-	    Xover_Headers[3].buf,      /* msgid */
-	    Xover_Headers[4].buf,      /* references */
+	    NNTP_Xover_Headers[0].buf,      /* subject */
+	    NNTP_Xover_Headers[1].buf,      /* from */
+	    NNTP_Xover_Headers[2].buf,      /* date */
+	    NNTP_Xover_Headers[3].buf,      /* msgid */
+	    NNTP_Xover_Headers[4].buf,      /* references */
 	    "",			       /* bytes */
-	    Xover_Headers[5].buf,      /* lines */
-	    Xover_Headers[6].buf       /* xref */
+	    NNTP_Xover_Headers[5].buf,      /* lines */
+	    NNTP_Xover_Headers[6].buf       /* xref */
 	    );
    return the_buf;
 }
 
-char *nntp_read_xover (char *the_buf, int len)
+static char *nntp_read_xover (char *the_buf, unsigned int len)
 {
    char buf[512];
-   int status, id;
+   int id;
    int already_tried = 0;
    
-   if (Slrn_NNTP_Has_XOver && !Suspend_XOver_For_Kill)
+   if (NNTP_Server_Obj.has_xover && !NNTP_Suspend_XOver_For_Kill)
      {
 	return nntp_read_line (the_buf, len);
      }
    
-   if (XOver_Next > XOver_Max) return NULL;
+   if (NNTP_XOver_Next > NNTP_XOver_Max) return NULL;
    
-   while (XOver_Done == 0)
+   while (NNTP_XOver_Done == 0)
      {
-	sprintf (buf, "HEAD %d", XOver_Next);
-	slrn_put_server (buf);
-	slrn_get_server (buf, sizeof(buf), 1);
-	if (atoi(buf) != OK_HEAD)
+	sprintf (buf, "HEAD %d", NNTP_XOver_Next);
+	if (OK_HEAD != slrn_put_server_cmd (buf, buf, sizeof(buf)))
 	  {
 	     already_tried = 0;
 	     /* This is ugly */
@@ -730,48 +781,43 @@ char *nntp_read_xover (char *the_buf, int len)
 	     
 	     do
 	       {
-		  slrn_put_server ("NEXT");
-		  slrn_get_server (buf, sizeof(buf), 1);
-		  if (atoi(buf) != OK_NOTEXT)
+		  switch (slrn_put_server_cmd ("NEXT", buf, sizeof (buf)))
 		    {
-		       if (atoi(buf) == ERR_NONEXT)
-			 {
-			    XOver_Done = 1;
-			    return NULL;
-			 }
+		     case ERR_NONEXT:
+		       NNTP_XOver_Done = 1;
+		       return NULL;
 		       
+		     case OK_NOTEXT:
+		       /* Ok, goto next article */
+		       break;
+		       
+		     default:
 		       slrn_exit_error ("Server failed NEXT request.");
 		    }
 		  
 		  id = atoi (buf + 4);
 		  /* Try going back 5 articles */
-		  if ((already_tried == 0) && (id < XOver_Next - 5))
+		  if ((already_tried == 0) && (id < NNTP_XOver_Next - 5))
 		    {
 		       
-		       sprintf (buf, "HEAD %d", XOver_Next - 5);
-		       slrn_put_server (buf);
-		       slrn_get_server (buf, sizeof(buf), 1);
+		       sprintf (buf, "HEAD %d", NNTP_XOver_Next - 5);
 		       
-		       if (atoi(buf) == OK_HEAD)
+		       if (OK_HEAD == slrn_put_server_cmd (buf, buf, sizeof(buf)))
 			 {
 			    /* read all header */
-			    while (nntp_read_line (buf, sizeof(buf)) != NULL)  ;
+			    nntp_discard_output (buf, sizeof (buf));
 			 }
 		       already_tried = 1;
 		    }
 	       }
-	     while (id < XOver_Next);
+	     while (id < NNTP_XOver_Next);
 	     
-	     slrn_put_server ("HEAD");
-	     slrn_get_server (buf, sizeof(buf), 1);
 	     
-	     if (atoi(buf) != OK_HEAD)
+	     if (OK_HEAD != slrn_put_server_cmd ("HEAD", buf, sizeof(buf)))
 	       {
 		  sprintf (buf, "HEAD %d", id);
-		  slrn_put_server (buf);
-		  slrn_get_server (buf, sizeof(buf), 1);
 		  
-		  if (atoi(buf) != OK_HEAD)
+		  if (OK_HEAD != slrn_put_server_cmd (buf, buf, sizeof(buf)))
 		    {
 		       already_tried++;
 		       if (already_tried > 10)
@@ -785,92 +831,95 @@ char *nntp_read_xover (char *the_buf, int len)
 	  }
 	/* extract article number */
 	id = atoi(buf + 4);
-	if (id >= XOver_Min) break;
+	if (id >= NNTP_XOver_Min) break;
 	
 	/* we just did a head.  Read it and do next until range is ok */
-	while ((0 == slrn_get_server (buf, sizeof(buf), 1)) && (*buf != '.'))
-	  {
-	     ;
-	  }
+	nntp_discard_output (buf, sizeof (buf));
 	
-	while (id < XOver_Min)
+	while (id < NNTP_XOver_Min)
 	  {
-	     slrn_put_server ("NEXT");
-	     slrn_get_server (buf, sizeof(buf), 1);
-	     
-	     status = atoi(buf);
-	     if (status == 223)
+	     if (OK_NOTEXT == slrn_put_server_cmd ("NEXT", buf, sizeof (buf)))
 	       {
 		  id = atoi(buf + 4);
 		  continue;
 	       }
-	     XOver_Done = 1;
+	     NNTP_XOver_Done = 1;
 	     break;
 	  }
      }
    
-   if (XOver_Done) return NULL;
-   if (id > XOver_Max)
+   if (NNTP_XOver_Done) return NULL;
+   if (id > NNTP_XOver_Max)
      {
 	/* clear out header text */
-	XOver_Done = 1;
+	NNTP_XOver_Done = 1;
 	while (nntp_read_line (buf, sizeof(buf)) != NULL)  ;
 	return NULL;
      }
    
    read_create_xover_line (id, the_buf);
    
-   if (id > XOver_Max)
+   if (id > NNTP_XOver_Max)
      {
-	XOver_Done = 1;
+	NNTP_XOver_Done = 1;
 	return NULL;
      }
    
    
-   slrn_put_server ("NEXT");
-   slrn_get_server (buf, sizeof(buf), 1);
-   
-   status = atoi(buf);
-   if (status == OK_NOTEXT)
+   if (OK_NOTEXT == slrn_put_server_cmd ("NEXT", buf, sizeof (buf)))
      {
-	XOver_Next = atoi(buf + 4);
-	if (XOver_Next > XOver_Max)
+	NNTP_XOver_Next = atoi(buf + 4);
+	if (NNTP_XOver_Next > NNTP_XOver_Max)
 	  {
-	     XOver_Done = 1;
+	     NNTP_XOver_Done = 1;
 	  }
      }
-   else XOver_Done = 1;
+   else NNTP_XOver_Done = 1;
    return the_buf;
 }
 
-void nntp_close_xover ()
+static void nntp_close_xover ()
 {
-   XOver_Done = 1;
+   NNTP_XOver_Done = 1;
 }
 
 
-char *nntp_head_from_msgid (char *msgid, char *buf, int len)
+static char *nntp_head_from_msgid (char *msgid, char *buf, unsigned int len)
 {
-   int id;
+   int id, status;
    
    if ((msgid == NULL) || (*msgid == 0)) return NULL;
+   
    sprintf (buf, "HEAD %s", msgid);
-   slrn_put_server (buf);
-   if (slrn_get_server (buf, len, 0) == -1)
+   status = put_server_with_reconnect (buf, buf, len);
+   if (OK_HEAD != status)
      {
-	slrn_error ("Server closed connection.");
+	if (ERR_FAULT == status)
+	  {
+	     slrn_error ("Server does not provide this capability.");
+	  }
 	return NULL;
      }
    
-   if (atoi (buf) != OK_HEAD) return NULL;
    id = atoi(buf + 4);
    if (id == 0) id = -1;
    return read_create_xover_line (id, buf);
 }
 
-int nntp_start_post (void)
+static int nntp_start_post (void)
 {
    char line[256];
+   /* Make sure we're connected to a server -- won't be the case the first
+    * time we post if we're reading news from a local spool.  In this case
+    * a late connect is better as we don't bother the server until we need
+    * to.
+    */
+   if (!NNTP_Server_Inited)
+     {
+	/* nntp_initialize_server gives an error if it fails, so we don't */
+	if (nntp_initialize_server () != 0)
+	  return -1;
+     }
    
    if (CONT_POST != put_server_with_reconnect ("POST", line, sizeof (line)))
      {
@@ -880,16 +929,16 @@ int nntp_start_post (void)
    return 0;
 }
 
-int nntp_end_post (void)
+static int nntp_end_post (void)
 {
    char line[256];
    
-#ifdef VMS
+#  ifdef VMS
    netwrite (SERVER_WRITE_FP, "\r\n.\r\n", 5);
-#else
+#  else
    fputs("\r\n.\r\n", SERVER_WRITE_FP);
    fflush (SERVER_WRITE_FP);
-#endif
+#  endif
    slrn_get_server (line, 255, 1);
    if (atoi(line) != 240)
      {
@@ -899,38 +948,38 @@ int nntp_end_post (void)
    return 0;
 }
 
-int nntp_printf (char *fmt, ...)
+static int nntp_printf (char *fmt, ...)
 {
    va_list ap;
-#ifdef VMS
+#  ifdef VMS
    char buf[512];
    va_start (ap, fmt);
    vsprintf (buf, fmt, ap);
    netwrite (SERVER_WRITE_FP, buf, strlen (buf));
-#else
+#  else
    va_start (ap, fmt);
    vfprintf (SERVER_WRITE_FP, fmt, ap);
-#endif
+#  endif
    return 0;
 }
 
-int nntp_puts (char *s)
+static int nntp_puts (char *s)
 {
-#ifndef VMS
+#  ifndef VMS
    fputs (s, SERVER_WRITE_FP);
-#else
+#  else
    netwrite (SERVER_WRITE_FP, s, strlen (s));
-#endif
+#  endif
    return 0;
 }
 
-int nntp_xhdr_command (char *hdr, int num, char *buf, unsigned int buflen)
+static int nntp_xhdr_command (char *hdr, int num, char *buf, unsigned int buflen)
 {
    char cmd[256], tmpbuf[1024];
    int found;
    unsigned int colon;
    
-   if (PROBE_XCMD(Has_Xhdr, "XHDR"))
+   if (PROBE_XCMD(NNTP_Has_Xhdr, "XHDR"))
      {
 	char *b, ch;
 	sprintf (cmd, "XHDR %s %d", hdr, num);
@@ -949,8 +998,7 @@ int nntp_xhdr_command (char *hdr, int num, char *buf, unsigned int buflen)
 	/* I should handle multi-line returns but I doubt that there will be
 	 * any for our use of xhdr
 	 */
-	while (nntp_read_line(tmpbuf, sizeof (tmpbuf)) != NULL)
-	  ;
+	nntp_discard_output (tmpbuf, sizeof (tmpbuf));
 	return 0;
      }
    
@@ -980,27 +1028,211 @@ int nntp_xhdr_command (char *hdr, int num, char *buf, unsigned int buflen)
 }
 
 
-int nntp_list_newsgroups (void)
+static int nntp_list_newsgroups (void)
 {
    char buf[256];
-   slrn_put_server ("LIST NEWSGROUPS");
-   slrn_get_server (buf, sizeof(buf), 1);
-   if (atoi (buf) != OK_GROUPS)
+   
+   if (OK_GROUPS != slrn_put_server_cmd ("LIST NEWSGROUPS", buf, sizeof (buf)))
      {
 	slrn_exit_error ("Server failed to return proper response to LIST NEWGROUPS");
      }
    return 0;
 }
 
-int nntp_list_active (void)
+static int nntp_list_active (void)
 {
    char buf[256];
-   slrn_put_server ("LIST");
-   slrn_get_server (buf, sizeof (buf), 1);
-   if (atoi(buf) != OK_GROUPS)
+   
+   if (OK_GROUPS != slrn_put_server_cmd ("LIST", buf, sizeof (buf)))
      {
 	slrn_exit_error("List failed: %s", buf);
      }
    return 0;
 }
 
+
+static char *nntp_getserverbyfile (char *file)
+{
+   static char buf[256];
+   register FILE *fp;
+   register char *cp;
+
+   if (NULL != (cp = getenv("NNTPSERVER")))
+     {
+	strcpy (buf, cp);
+	return buf;
+     }
+   
+   if (file == NULL)
+     {
+#ifdef NNTPSERVER_NAME
+	strcpy (buf, NNTPSERVER_NAME);
+	return buf;
+#else
+	return NULL;
+#endif
+     }
+   
+   fp = fopen(file, "r");
+   if (fp == NULL)
+     return NULL;
+   
+   while (fgets(buf, sizeof (buf), fp) != NULL)
+     {
+	if (*buf == '\n' || *buf == '#')
+	  continue;
+	cp = slrn_strchr (buf, '\n');
+	if (cp)
+	  *cp = '\0';
+	(void) fclose(fp);
+	return buf;
+     }
+   
+   (void) fclose(fp);
+   return NULL;    /* No entry */
+}
+
+static int nntp_init_objects (void)
+{
+   NNTP_Post_Obj.start = nntp_start_post;
+   NNTP_Post_Obj.end = nntp_end_post;
+   NNTP_Post_Obj.printf = nntp_printf;
+   NNTP_Post_Obj.puts = nntp_puts;
+   NNTP_Post_Obj.can_post = 1;
+   
+   NNTP_Server_Obj.select_group = nntp_select_group;
+   NNTP_Server_Obj.read_line = nntp_read_line;
+   NNTP_Server_Obj.close = nntp_close_server;
+   NNTP_Server_Obj.initialize = nntp_initialize_server;
+   NNTP_Server_Obj.select_article = nntp_select_article;
+   NNTP_Server_Obj.head_from_msgid = nntp_head_from_msgid;
+   NNTP_Server_Obj.read_xover = nntp_read_xover;
+   NNTP_Server_Obj.open_xover = nntp_open_xover;
+   NNTP_Server_Obj.close_xover = nntp_close_xover;
+   NNTP_Server_Obj.put_server_cmd = slrn_put_server_cmd;
+   NNTP_Server_Obj.xpat_cmd = nntp_xpat_cmd;
+   NNTP_Server_Obj.xhdr_command = nntp_xhdr_command;
+   NNTP_Server_Obj.get_extra_xover_header = nntp_get_extra_xover_header;
+   NNTP_Server_Obj.close_suspend_xover = nntp_close_suspend_xover;
+   NNTP_Server_Obj.open_suspend_xover = nntp_open_suspend_xover;
+   NNTP_Server_Obj.xgtitle_cmd = nntp_xgtitle_cmd;
+   NNTP_Server_Obj.has_xover = 0;
+   NNTP_Server_Obj.has_cmd = nntp_has_cmd;
+   NNTP_Server_Obj.list_newsgroups = nntp_list_newsgroups;
+   NNTP_Server_Obj.list_active = nntp_list_active;
+   
+   return 0;
+}
+
+
+static char *nntp_port_change (char *hostname)
+{
+   char *colon;
+   char *h;
+   
+   colon = slrn_strchr (hostname, ':');
+   if (colon == NULL) return hostname;
+   
+   h = SLmake_string (hostname);
+   if (h == NULL)
+     slrn_exit_error ("Out of memory.");
+   
+   colon = h + (colon - hostname);
+   *colon++ = 0;
+   Slrn_NNTP_Port = atoi (colon);
+   return h;
+}
+
+#ifndef NNTPSERVER_FILE
+# define NNTPSERVER_FILE NULL
+#endif
+
+static int nntp_get_servername (void)
+{
+   char *host = Slrn_NNTP_Server_Name;
+   
+   if ((host == NULL)
+       && (NULL == (host = nntp_getserverbyfile(NNTPSERVER_FILE))))
+     {
+	fprintf (stderr, "You need to set the NNTPSERVER environment variable to your server name.\r\n");
+#ifdef VMS
+	fprintf (stderr, "Example: $ define/job NNTPSERVER my.news.server\r\n");
+#else
+# ifdef __os2__
+	fprintf (stderr, "Example: set NNTPSERVER=my.news.server\r\n");
+# else
+	fprintf (stderr, "Example (csh): setenv NNTPSERVER my.news.server\r\n");
+# endif
+#endif
+	return -1;
+     }
+
+   Slrn_NNTP_Server_Name = nntp_port_change (host);
+   
+   if (NNTP_Server_Obj.name != NULL)
+     SLFREE (NNTP_Server_Obj.name);
+   
+   NNTP_Server_Obj.name = slrn_make_startup_string (host);
+   return 0;
+}
+
+   
+
+static int nntp_select_server_object (void)
+{
+   Slrn_Server_Obj = &NNTP_Server_Obj;
+   
+   if (NNTP_Server_Obj.name == NULL)
+     return nntp_get_servername ();
+   
+   return 0;
+}
+
+static int nntp_select_post_object (void)
+{
+   Slrn_Post_Obj = &NNTP_Post_Obj;
+   
+   if (NNTP_Server_Obj.name == NULL)
+     return nntp_get_servername ();
+   
+   return 0;
+}
+
+static void nntp_usage (void)
+{
+   fputs ("--nntp options:\n\
+-h nntp-host    Host name to connect to.  This overrides NNTPSERVER variable.\n\
+-p NNTP-PORT    Set the NNTP port to NNTP-PORT. The default value is 119.\n\
+                 Note: This option has no effect on some systems.\n\
+",
+	  stdout);
+   exit (0);
+}   
+/* returns number parsed */
+static int nntp_parse_args (char **argv, int argc)
+{
+   int i;
+   
+   for (i = 0; i < argc; i++)
+     {
+	if (!strcmp (argv[i], "--help"))
+	  nntp_usage ();
+	else if (i + 1 < argc)
+	  {
+	     if (!strcmp ("-p", argv[i]))
+	       {
+		  Slrn_NNTP_Port = atoi (argv[++i]);
+	       }
+	     else if (!strcmp ("-h", argv[i]))
+	       {
+		  Slrn_NNTP_Server_Name = argv[++i];
+	       }
+	     else break;
+	  }
+	else break;
+     }
+   
+   return i;
+}
+
+	     
