@@ -1,4 +1,5 @@
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -12,7 +13,7 @@
 #include "art.h"
 #include "score.h"
 #include "misc.h"
-#include "nntp.h"
+#include "server.h"
 
 /* This file will implement article scoring for slrn.  Basically the idea is
  * that the user will provide a set of regular expressions that will act on
@@ -100,19 +101,29 @@ typedef struct Score_Type
 }
 Score_Type;
 
-typedef struct Group_Score_Type
+#define MAX_GROUP_REGEXP_SIZE	256
+#define MAX_GROUP_NAME_LEN	80
+
+typedef struct Group_Score_Name_Type
 {
    SLRegexp_Type group_regexp;
-   char name[80];		       /* group name or pattern */
-   unsigned char buf[256];	       /* for compiled pattern */
+   char name[MAX_GROUP_NAME_LEN];      /* group name or pattern */
+   unsigned char buf[MAX_GROUP_REGEXP_SIZE];/* for compiled pattern */
+   struct Group_Score_Name_Type *next;
+}
+Group_Score_Name_Type;
+
+typedef struct Group_Score_Type
+{
+   Group_Score_Name_Type gsnt;
    PScore_Type *pst;
    struct Group_Score_Type *next;
 }
 Group_Score_Type;
 
-Group_Score_Type *Group_Score_Root;
-Group_Score_Type *Group_Score_Next;
-Score_Type *Score_Root, *Score_Tail;
+static Group_Score_Type *Group_Score_Root;
+static Group_Score_Type *Group_Score_Next;
+static Score_Type *Score_Root, *Score_Tail;
 
 int Slrn_Apply_Score = 1;
 
@@ -171,7 +182,7 @@ int slrn_score_header (Slrn_Header_Type *h, char *newsgroup)
 		  if (Slrn_Score_After_XOver)
 		    {
 		       if (((Slrn_Perform_Scoring & SLRN_EXPENSIVE_SCORING) == 0)
-			   || (-1 == nntp_xhdr_command (srt->generic_keyword,
+			   || (-1 == Slrn_Server_Obj->xhdr_command (srt->generic_keyword,
 							h->number, buf, sizeof(buf))))
 			 s = NULL;
 		       else s = buf;
@@ -179,7 +190,7 @@ int slrn_score_header (Slrn_Header_Type *h, char *newsgroup)
 		  else
 		    {
 		       if (Slrn_Perform_Scoring & SLRN_EXPENSIVE_SCORING)
-			 s = nntp_get_extra_xover_header (srt->generic_keyword);
+			 s = Slrn_Server_Obj->get_extra_xover_header (srt->generic_keyword);
 		       else s = NULL;
 		    }
 		  break;
@@ -406,16 +417,26 @@ int slrn_open_score (char *group_name)
    generic = 0;
    while (gsc != NULL)
      {
-	SLRegexp_Type *re = &gsc->group_regexp;
+	Group_Score_Name_Type *gsnt;
+	SLRegexp_Type *re;
 	
-	if ((re->min_length <= n)
-	    && (NULL != SLang_regexp_match ((unsigned char *) group_name, n, re)))
+	gsnt = &gsc->gsnt;
+	
+	while (gsnt != NULL)
 	  {
-	     if (-1 == chain_group_regexp (gsc->pst, &generic))
+	     re = &gsnt->group_regexp;
+	
+	     if ((re->min_length <= n)
+		 && (NULL != SLang_regexp_match ((unsigned char *) group_name, n, re)))
 	       {
-		  free_group_chain ();
-		  return -1;
+		  if (-1 == chain_group_regexp (gsc->pst, &generic))
+		    {
+		       free_group_chain ();
+		       return -1;
+		    }
+		  break;
 	       }
+	     gsnt = gsnt->next;
 	  }
 	gsc = gsc->next;
      }
@@ -428,10 +449,10 @@ int slrn_open_score (char *group_name)
        || (0 == (Slrn_Perform_Scoring & SLRN_EXPENSIVE_SCORING)))
      return 1;
    
-   if ((generic >= 1) || !Slrn_NNTP_Has_XOver || !nntp_has_cmd ("XHDR"))
+   if ((generic >= 1) || !Slrn_Server_Obj->has_xover || !Slrn_Server_Obj->has_cmd ("XHDR"))
      {
 	Slrn_Score_After_XOver = 0;
-	nntp_open_suspend_xover ();
+	Slrn_Server_Obj->open_suspend_xover ();
      }
    return 1;
 }
@@ -439,7 +460,7 @@ int slrn_open_score (char *group_name)
 void slrn_close_score (void)
 {
    free_group_chain ();
-   nntp_close_suspend_xover ();
+   Slrn_Server_Obj->close_suspend_xover ();
 }
 
 
@@ -497,6 +518,63 @@ static int add_group_regexp (PScore_Type *pst,
    return 0;
 }
 
+
+static int compile_group_names (char *group, Group_Score_Type *gst)
+{
+   char *comma;
+   Group_Score_Name_Type *gsnt;
+   unsigned int num_processed = 0;
+   
+   gsnt = &gst->gsnt;
+   
+   comma = group;
+   while (comma != NULL)
+     {
+	SLRegexp_Type *re;
+	
+	group = slrn_skip_whitespace (group);
+	
+	comma = slrn_strchr ((char *) group, ',');
+	if (comma != NULL) 
+	  {
+	     *comma++ = 0;
+	  }
+	
+	(void) slrn_trim_string (group);
+	if (*group == 0) continue;
+	
+	strncpy (gsnt->name, group, MAX_GROUP_NAME_LEN - 1);
+	/* Note: because of the memset, this string is null terminated. */
+   
+	re = &gsnt->group_regexp;
+	re->pat = (unsigned char *) slrn_fix_regexp ((char *)group);
+	re->buf = gsnt->buf;
+	re->buf_len = MAX_GROUP_REGEXP_SIZE;
+	re->case_sensitive = 0;
+	if (0 != SLang_regexp_compile (re))
+	  {
+	     return -1;
+	  }
+	
+	if (comma != NULL)
+	  {
+	     Group_Score_Name_Type *gsnt1;
+	     
+	     gsnt1 = (Group_Score_Name_Type *) SLMALLOC (sizeof (Group_Score_Name_Type));
+	     if (gsnt1 == NULL) 
+	       slrn_exit_error ("Memory allocation failure.");
+	     memset ((char *) gsnt1, 0, sizeof (Group_Score_Name_Type));
+	     
+	     gsnt->next = gsnt1;
+	     gsnt = gsnt1;
+	     group = comma;
+	  }
+	num_processed++;
+     }
+   if (num_processed) return 0;
+   return -1;
+}
+
 static PScore_Type *create_new_score (unsigned char *group, int new_group_flag,
 				      int score, unsigned int pscore_flags)
 {
@@ -505,12 +583,12 @@ static PScore_Type *create_new_score (unsigned char *group, int new_group_flag,
    if (new_group_flag)
      {
 	Group_Score_Type *gst, *tail;
-	SLRegexp_Type *re;
 	
 	tail = Group_Score_Next = Group_Score_Root;
 	while (Group_Score_Next != NULL)
 	  {
-	     if (!strcmp ((char *) group, Group_Score_Next->name))
+	     if ((Group_Score_Next->gsnt.next == NULL)
+		 && !strcmp ((char *) group, Group_Score_Next->gsnt.name))
 	       break;
 	     tail = Group_Score_Next;
 	     Group_Score_Next = Group_Score_Next->next;
@@ -524,19 +602,9 @@ static PScore_Type *create_new_score (unsigned char *group, int new_group_flag,
 	     
 	     memset ((char *)gst, 0, sizeof (Group_Score_Type));
 	     
-	     strncpy (gst->name, (char *) group, sizeof (gst->name) - 1);
-	     /* Note: necause of the memset, this string is null terminated. */
-	     
-	     re = &gst->group_regexp;
-	     re->pat = (unsigned char *) slrn_fix_regexp ((char *)group);
-	     re->buf = gst->buf;
-	     re->buf_len = sizeof (gst->buf);
-	     re->case_sensitive = 0;
-	     if (0 != SLang_regexp_compile (re))
-	       {
-		  return NULL;
-	       }
-	     
+	     if (-1 == compile_group_names ((char *)group, gst))
+	       return NULL;
+
 	     if (Group_Score_Root == NULL)
 	       {
 		  Group_Score_Root = gst;
@@ -583,6 +651,15 @@ static void free_group_scores (void)
      {
 	Group_Score_Type *gnext = Group_Score_Root->next;
 	PScore_Type *pst = Group_Score_Root->pst;
+	Group_Score_Name_Type *gsnt;
+	
+	gsnt = Group_Score_Root->gsnt.next;
+	while (gsnt != NULL)
+	  {
+	     Group_Score_Name_Type *next = gsnt->next;
+	     SLFREE (gsnt);
+	     gsnt = next;
+	  }
 	
 	while (pst != NULL)
 	  {
@@ -602,6 +679,7 @@ static void free_group_scores (void)
 	     SLFREE (pst);
 	     pst = pnext;
 	  }
+	
 	SLFREE (Group_Score_Root);
 	Group_Score_Root = gnext;
      }
@@ -663,7 +741,7 @@ int slrn_read_score_file (char *name)
 	free_group_scores ();
      }
    
-   fp = slrn_open_home_file (name, "rt", file, 0);
+   fp = slrn_open_home_file (name, "r", file, 0);
    if (fp == NULL)
      {
 	Slrn_Apply_Score = 0;
@@ -822,7 +900,7 @@ int slrn_edit_score (Slrn_Header_Type *h, char *newsgroup)
    unsigned int mm = 0, dd = 0, yy = 0;
    int use_expired = 0;
    unsigned int linenum = 0;
-   char *q;
+   char *q, *ng = newsgroup;
    int score;
    FILE *fp;
    time_t myclock;
@@ -845,6 +923,9 @@ int slrn_edit_score (Slrn_Header_Type *h, char *newsgroup)
 	  return -1;
 	if (score == 0) return -1;
 	
+	if ('a' == (0x20 | slrn_get_response ("TtaA", "Which newsgroups: T-his group, A-ll groups")))
+	  ng = "*";
+
 	while (1)
 	  {
 	     *qregexp = 0;
@@ -869,8 +950,8 @@ int slrn_edit_score (Slrn_Header_Type *h, char *newsgroup)
 	  }
      }
    
-   if ((NULL == (fp = slrn_open_home_file (Slrn_Score_File, "r+t", file, 1)))
-       && (NULL == (fp = slrn_open_home_file (Slrn_Score_File, "w+t", file, 1))))
+   if ((NULL == (fp = slrn_open_home_file (Slrn_Score_File, "r+", file, 1)))
+       && (NULL == (fp = slrn_open_home_file (Slrn_Score_File, "w+", file, 1))))
      {
 	slrn_error ("Unable to open %s", file);
 	return -1;
@@ -885,8 +966,8 @@ int slrn_edit_score (Slrn_Header_Type *h, char *newsgroup)
 	  }
 	
 	myclock = time((time_t *) 0);
-	fprintf (fp, "\n%%Score created by slrn on %s[%s]\nScore: %d\n",
-		 (char *) ctime(&myclock), newsgroup, score);
+	fprintf (fp, "\n%%Score created by slrn on %s\n[%s]\nScore: %d\n",
+		 (char *) ctime(&myclock), ng, score);
 	
 	if (use_expired)
 	  fprintf (fp, "Expires: %u/%u/%u\n", mm, dd, yy);
