@@ -1,7 +1,8 @@
 /*  Copyright (c) 1995 John E. Davis (davis@space.mit.edu)
  *  All rights reserved.
  */
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #include <signal.h>
@@ -38,10 +39,9 @@
 
 #include <slang.h>
 
+#include <errno.h>
 
-
-#include "clientlib.h" 
-#include "nntp.h"
+#include "server.h"
 #include "slrn.h"
 #include "group.h"
 #include "misc.h"
@@ -54,21 +54,17 @@
 # include <os2.h>
 #endif
 
-#ifndef NNTPSERVER_FILE
-# define NNTPSERVER_FILE NULL
-#endif
-
 #ifdef SIGSTOP
 # define USE_RESUME_SMG
 #endif
 
-
 int Slrn_TT_Initialized = 0;
 
 /* If -1, force mouse.  If 1 the mouse will be used on in XTerm.  If 0, 
  * do not use it.
  */
 int Slrn_Use_Mouse;
+char *Slrn_Version = SLRN_VERSION;
 
 int Slrn_Suspension_Ok;
 void slrn_set_suspension (int ok)
@@ -84,67 +80,174 @@ void slrn_smg_refresh (void)
 {
    int ok = Slrn_Suspension_Ok;
    if (ok) slrn_set_suspension (0);
-   SLsmg_refresh ();
+   if (Slrn_TT_Initialized & 1) SLsmg_refresh ();
    if (ok) slrn_set_suspension (1);
 }
 
+static void set_signal_to_restart_syscalls (int sig)
+{
+#ifdef SLRN_POSIX_SIGNALS
+   struct sigaction sa;
+   
+   if (-1 == sigaction (sig, NULL, &sa)) return;
+   
+# ifdef SIGALRM
+   if (sig == SIGALRM)
+     {
+#  ifdef SA_INTERRUPT
+	sa.sa_flags |= SA_INTERRUPT;
+#  endif
+     }
+   else
+     {
+# endif				       /* SIGALRM */
+# ifdef SA_RESTART
+	sa.sa_flags |= SA_RESTART;
+# endif
+# ifdef SIGALRM
+     }
+# endif
+   
+   (void) sigaction (sig, &sa, NULL);
+
+#endif				       /* SLRN_POSIX_SIGNALS */
+}
+
+
+
+static int Can_Suspend;
+
 void slrn_suspend (int sig)
 {
-#ifdef SIGSTOP
+#ifndef SIGSTOP
+   slrn_error ("Not implemented.");
+#else
+   
    int init = Slrn_TT_Initialized;
+# ifdef SLRN_POSIX_SIGNALS
+   sigset_t mask;
+
+   if (Can_Suspend == 0)
+     {
+	slrn_error ("Suspension not allowed by shell.");
+	return;
+     }
+   
+   sigemptyset (&mask);
+   sigaddset (&mask, SIGTSTP);
+   
+   /* This function resets SIGTSTP to default */
+   slrn_reset_display (sig);
+   
+   kill (getpid (), SIGTSTP);
+   
+   /* If SIGTSTP is pending, it will be delivered now.  That's ok. */
+   sigprocmask (SIG_UNBLOCK, &mask, NULL);
+# else
+   
+   if (Can_Suspend == 0)
+     {
+	slrn_error ("Suspension not allowed by shell.");
+	return;
+     }
+   
    slrn_reset_display (sig);
    kill(getpid(),SIGSTOP);
-   slrn_init_display (init, sig);
-# ifdef USE_RESUME_SMG
-   if (Slrn_Suspension_Ok) SLtty_set_suspend_state (1);
 # endif
-#else
-   slrn_error ("Not implemented.");
+   
+   slrn_init_display (init, sig);
+   
+   if (Slrn_Suspension_Ok) SLtty_set_suspend_state (1);
 #endif
 }
 
+static void test_lock( char *file )
+{
+   int pid;
+   FILE *fp;
+   
+   if ((fp = fopen (file, "r")) != NULL)
+     {
+	if (1 == fscanf (fp, "%d", &pid) )
+	  {
+	     if ((pid > 0) && (0 == kill (pid, 0)))
+	       {
+		  slrn_exit_error ("slrn: pid %d is locking the newsrc file.", pid);
+	       }
+	  }
+	slrn_fclose (fp);
+     }
+}
+
+static int make_lock( char *file )
+{
+   int pid;
+   FILE *fp;
+   
+#ifdef VMS
+   fp = fopen (file, "w", "fop=cif");
+#else
+   fp = fopen (file, "w");
+#endif
+   
+   if (fp == NULL) return -1;
+   
+   pid = getpid ();
+   fprintf (fp, "%d", pid);
+   slrn_fclose (fp);
+   return 0;
+}
+
+
 static void lock_file (int how)
 {
    char file[1024];
    char name[256];
-   char buf[80];
    static int not_ok_to_unlock;
-   FILE *fp;
+#if SLRN_HAS_RNLOCK
+   int rnlock = 0;
+   char file_rn[1024];
+#endif
    
    if (Slrn_Newsrc_File == NULL) return;
    if (not_ok_to_unlock) return;
    
    not_ok_to_unlock = 1;
+
+#ifndef __os2__
    sprintf (name, "%s-lock", Slrn_Newsrc_File);
+#else
+   sprintf (name, "lk-%s", Slrn_Newsrc_File);
+#endif
+     
    slrn_make_home_filename (name, file);
    
+#if SLRN_HAS_RNLOCK
+   if (0 == strcmp(".newsrc", Slrn_Newsrc_File))
+     {
+	rnlock = 1;
+	slrn_make_home_filename (".rnlock", file_rn);
+     }
+#endif
+
    if (how == 1)
      {
-	int pid;
-	if ((fp = fopen (file, "r")) != NULL)
-	  {
-	     if (NULL != fgets (buf, 79, fp))
-	       {
-		  sscanf (buf, "%d", &pid);
-		  if ((pid > 0) && (0 == kill (pid, 0)))
-		    {
-		       slrn_exit_error ("slrn: pid %d is locking the newsrc file.", pid);
-		    }
-	       }
-	     slrn_fclose (fp);
-	  }
-#ifdef VMS
-	fp = fopen (file, "w", "fop=cif");
-#else
-	fp = fopen (file, "w");
+	test_lock (file);
+#if SLRN_HAS_RNLOCK
+	if (rnlock) test_lock (file_rn);
 #endif
-	if (NULL == fp)
+	if (-1 == make_lock (file))
 	  {
 	     slrn_exit_error ("Unable to create lock file %s.", file);
 	  }
-	pid = getpid ();
-        fprintf (fp, "%d", pid);
-	slrn_fclose (fp);
+	
+#if SLRN_HAS_RNLOCK
+	if (rnlock && (-1 == make_lock (file_rn)))
+	  {
+	     slrn_delete_file (file); /* delete the "normal" lock file */
+	     slrn_exit_error ("Unable to create lock file %s.", file_rn);
+	  }
+#endif       
      }
    else
      {
@@ -152,6 +255,12 @@ static void lock_file (int how)
 	  {
 	     /* slrn_exit_error ("Unable to remove lockfile %s.", file); */
 	  }
+#if SLRN_HAS_RNLOCK
+        if (rnlock && -1 == slrn_delete_file (file_rn))
+	  {
+	     /* slrn_exit_error ("Unable to remove lockfile %s.", file_rn); */
+	  }
+#endif
      }
    not_ok_to_unlock = 0;
 }
@@ -173,13 +282,18 @@ static void slrn_set_screen_size (int sig)
 #ifdef TIOCGWINSZ
    struct winsize wind_struct;
    
-   if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
-       || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
-       || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
+   do
      {
-	c = (int) wind_struct.ws_col;
-	r = (int) wind_struct.ws_row;
+	if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
+	    || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
+	    || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
+	  {
+	     c = (int) wind_struct.ws_col;
+	     r = (int) wind_struct.ws_row;
+	     break;
+	  }
      }
+   while (errno == EINTR);
    
 #endif
 #ifdef VMS
@@ -222,7 +336,7 @@ static void slrn_set_screen_size (int sig)
    SLtt_Screen_Rows = r;
    SLtt_Screen_Cols = c;
    
-   Group_Window_Size = SLtt_Screen_Rows - 3;
+   Slrn_Group_Window_Size = SLtt_Screen_Rows - 3;
    
 #ifdef SIGWINCH
    if (sig == SIGWINCH)
@@ -237,11 +351,85 @@ static void slrn_set_screen_size (int sig)
 }
 
 
-void (*Slrn_Hangup_Hook) (void);
+void (*Slrn_Hangup_Hook) (int);
+
 static void slrn_hangup (int sig)
 {
-   if (Slrn_Hangup_Hook != NULL) (*Slrn_Hangup_Hook) ();
-   exit (-1);
+   slrn_init_hangup_signals (0);
+   if (Slrn_Hangup_Hook != NULL) (*Slrn_Hangup_Hook) (sig);
+   else
+     slrn_write_newsrc ();
+   
+   slrn_quit (sig);
+}
+
+
+static void init_like_signals (int argc, int *argv, 
+			       void (*f0)(int), 
+			       void (*f1)(int), 
+			       int state)
+{
+#ifdef HAVE_SIGACTION
+   struct sigaction sa;
+#endif
+   int i;
+
+   if (state == 0)
+     {
+	for (i = 0; i < argc; i++)
+	  signal (argv[i], f0);
+	return;
+     }
+
+   for (i = 0; i < argc; i++)
+     {
+	int sig = argv[i];
+	signal (sig, f1);
+
+#if defined(SLRN_POSIX_SIGNALS)
+	if (-1 != sigaction (sig, NULL, &sa))
+	  {
+	     int j;
+	     for (j = 0; j < argc; j++)
+	       {
+		  if (j != i) sigaddset (&sa.sa_mask, argv[j]);
+	       }
+	     
+	     (void) sigaction (sig, &sa, NULL);
+	  }
+#endif
+     }
+}
+
+static void init_suspend_signals (int state)
+{
+   int argv[2];
+   int argc = 0;
+   
+#ifdef SIGTSTP
+   argv[argc++] = SIGTSTP;
+#endif
+   
+#ifdef SIGTTIN
+   argv[argc++] = SIGTTIN;
+#endif
+
+   init_like_signals (argc, argv, SIG_DFL, slrn_suspend, state);
+}
+
+void slrn_init_hangup_signals (int state)
+{
+   int argv[2];
+   int argc = 0;
+
+#ifdef SIGHUP
+   argv[argc++] = SIGHUP;
+#endif
+#ifdef SIGTERM
+   argv[argc++] = SIGTERM;
+#endif
+   
+   init_like_signals (argc, argv, SIG_IGN, slrn_hangup, state);
 }
 
 void slrn_enable_mouse (int mode)
@@ -255,14 +443,13 @@ void slrn_enable_mouse (int mode)
 #endif
 }
 
+
 /* if 1, init just slsmg.  If 2, just sltt, if 3, do both */
 void slrn_init_display (int init_mode, int sig)
 {
    Slrn_TT_Initialized |= init_mode;
    
-#ifdef SIGTSTP
-   signal(SIGTSTP, slrn_suspend);
-#endif
+   init_suspend_signals (1);
    
 #ifdef SIGPIPE
    signal(SIGPIPE, SIG_IGN);
@@ -271,15 +458,16 @@ void slrn_init_display (int init_mode, int sig)
    /* Allow background writes */
    signal (SIGTTOU, SIG_IGN);
 #endif
-   signal (SIGHUP, slrn_hangup);
-#ifdef SIGTERM
-   signal (SIGTERM, slrn_hangup);
-#endif
+   
+   slrn_init_hangup_signals (1);
    
    if (init_mode & 2)
      {
 	SLang_init_tty (7, 1, 0);
 	SLang_set_abort_signal (NULL);
+#ifdef SIGINT
+	set_signal_to_restart_syscalls (SIGINT);
+#endif
      }
    
    if (init_mode & 1)
@@ -329,15 +517,11 @@ void slrn_reset_display (int sig)
    
    if (init & 2) SLang_reset_tty ();
    
-#ifdef SIGTSTP
-   signal(SIGTSTP, SIG_DFL);
-#endif
+   init_suspend_signals (0);
    
    Slrn_TT_Initialized = 0;
 }
 
-char *Slrn_Server_Name;
-
 void slrn_tty_message (int newline_flag, char *fmt, ...)
 {
    va_list ap;
@@ -358,7 +542,7 @@ int slrn_get_new_news (int no_new_groups, int create_flag)
    char msg3[] = "Checking news via active file...";
    char *msg;
    
-   if (nntp_initialize_server (Slrn_Server_Name) != 0) return (-1);
+   if (Slrn_Server_Obj->initialize () != 0) return (-1);
    
    if (create_flag == 0)
      {
@@ -396,13 +580,87 @@ int slrn_get_new_news (int no_new_groups, int create_flag)
 char *Slrn_Newsrc_File = NULL;
 char *Slrn_Lib_Dir = NULL;
 
+#if !defined (SLRN_DEFAULT_SERVER_OBJ)
+# if SLRN_HAS_NNTP_SUPPORT
+#  define SLRN_DEFAULT_SERVER_OBJ SLRN_SERVER_ID_NNTP
+# else
+#  define SLRN_DEFAULT_SERVER_OBJ SLRN_SERVER_ID_SPOOL
+# endif
+#endif
+
+#if !defined (SLRN_DEFAULT_POST_OBJ)
+# if SLRN_HAS_NNTP_SUPPORT
+#  define SLRN_DEFAULT_POST_OBJ SLRN_POST_ID_NNTP
+# else
+#  define SLRN_DEFAULT_POST_OBJ SLRN_POST_ID_INEWS
+# endif
+#endif
+
+
+static void version (void)
+{
+   unsigned int a, b, c;
+   
+   a = SLANG_VERSION/10000;
+   b = (SLANG_VERSION  - a * 10000) / 100;
+   c = SLANG_VERSION - (a * 10000) - (b * 100);
+   
+#if defined(__DATE__) && defined(__TIME__)
+   fprintf (stdout, "slrn version: %s (%s %s)\n", Slrn_Version, __DATE__, __TIME__);
+#else
+   fprintf (stdout, "slrn version: %s\n", Slrn_Version);
+#endif
+   fprintf (stdout, "S-Lang version: %u.%u.%u\n", a, b, c);
+   
+   fprintf (stdout, "\nslrn compiled with support for:\n");
+#if SLRN_HAS_NNTP_SUPPORT
+   fputs (" nntp", stdout);
+#endif
+#if SLRN_HAS_SPOOL_SUPPORT
+   fputs (" spool", stdout);
+#endif
+#if SLRN_HAS_NNTP_SUPPORT
+   fputs (" inews", stdout);
+#endif
+   
+   fputs ("\nDefault server object: ", stdout);
+   switch (SLRN_DEFAULT_SERVER_OBJ)
+     {
+      case SLRN_SERVER_ID_NNTP:
+	fputs ("nntp", stdout);
+	break;
+
+      case SLRN_SERVER_ID_SPOOL:
+	fputs ("spool", stdout);
+	break;
+     }
+   
+   fputs ("\nDefault posting mechanism: ", stdout);
+   switch (SLRN_DEFAULT_POST_OBJ)
+     {
+      case SLRN_POST_ID_NNTP:
+	fputs ("nntp", stdout);
+	break;
+
+      case SLRN_POST_ID_INEWS:
+	fputs ("inews", stdout);
+#if SLRN_FORCE_INEWS
+	fputs (" (fixed)", stdout);
+#endif
+	break;
+     }
+   
+   fputc ('\n', stdout);
+   exit (0);
+}
+
+   
 static void usage (char *extra)
 {
-   fprintf (stderr, "slrn (%s):\n\
-Usage: slrn [-n] [-C] [-h nntp-host]  [-f newsrc-file] [-create]\n\
+   fputs ("\
+Usage: slrn [--inews ...] [--nntp ...] [--spool ...] OPTIONS\n\
 -n              Do not check for new groups.  This usually results in\n\
                  a faster startup.\n\
--h nntp-host    Host name to connect to.  This overrides NNTPSERVER variable.\n\
 -f newsrc-file  Name of the newsrc file to use.\n\
 -C              Use colors.\n\
 -create         Create a newsrc file by getting list of groups from server.\n\
@@ -411,44 +669,60 @@ Usage: slrn [-n] [-C] [-h nntp-host]  [-f newsrc-file] [-create]\n\
                  The resulting file can be several hundred Kilobytes!\n\
 -i init-file    Name of initialization file to use (default .slrnrc)\n\
 -Dname          Add 'name' to list of predefined preprocessing tokens.\n\
--p NNTP-PORT    Set the NNTP port to NNTP-PORT. The default value is 119.\n\
-                 Note: This option has no effect on some systems.\n\
 -k              Do not process score file.\n\
 -k0             Process score file but inhibit expensive scores.\n\
 -a              Use active file for getting new news\n\
 -m              Force XTerm mouse reporting\n\
--help           Print this usage.\n",
-	    Slrn_Version);
+--help          Print this usage.\n\
+--version       Show version\n\
+\n\
+For additional info use one of forms:\n\
+    slrn --inews --help\n\
+    slrn --nntp --help\n\
+    slrn --spool --help\n\
+", 
+	  stderr);
+
    
    if (extra != NULL)
      {
 	fprintf (stderr, "\n%s\n", extra);
      }
-   
-   exit (-1);
+   exit (1);
 }
 
-static char *port_change (char *hostname)
+
+
+static int parse_object_args (char *obj, char **argv, int argc)
 {
-   char *colon;
-   char *h;
+   int num_parsed;
+   int zero_ok = 1;
    
-   colon = slrn_strchr (hostname, ':');
-   if (colon == NULL) return hostname;
+   if (obj == NULL)
+     {
+	zero_ok = 0;
+#if SLRN_DEFAULT_SERVER_OBJ == SLRN_SERVER_ID_SPOOL
+	obj = "spool";
+#endif
+#if SLRN_DEFAULT_SERVER_OBJ == SLRN_SERVER_ID_NNTP
+	obj = "nntp";
+#endif
+     }
    
-   h = SLmake_string (hostname);
-   if (h == NULL)
-     slrn_exit_error ("Out of memory.");
+   num_parsed = slrn_parse_object_args (obj, argv, argc);
+   if (num_parsed == -1)
+     {
+	slrn_exit_error ("%s is not a supported option.", *argv);
+     }
+   if ((num_parsed == 0) && (zero_ok == 0))
+     usage (NULL);
    
-   colon = h + (colon - hostname);
-   *colon++ = 0;
-   Slrn_NNTP_Port = atoi (colon);
-   return h;
+   return num_parsed;
 }
 
 int main (int argc, char **argv)
 {
-   char *host = NULL, *hlp_file;
+   char *hlp_file;
    int i;
    int create_flag = 0;
    int no_new_groups = 0;
@@ -456,34 +730,58 @@ int main (int argc, char **argv)
    int use_color = 0;
    int use_mouse = 0;
    int dsc_flag = 0;
+   int use_active = 0;
    FILE *fp;
    char file[512];
+   char *init_file = SLRN_USER_SLRNRC_FILENAME;
+   
+   Can_Suspend = 0;
+#ifdef SIGTSTP
+   if (SIG_DFL == signal (SIGTSTP, SIG_DFL))
+     {     
+	Can_Suspend = 1;
+     }
+#endif
+   
 #ifdef __os2__
-   char *init_file = "slrn.rc";
+   SLdefine_for_ifdef ("OS2");
 #else
-   char *init_file = ".slrnrc";
-#endif
-   
-   
-#ifdef unix
-   SLdefine_for_ifdef ("UNIX");
-#endif
-#ifdef VMS
+# ifdef VMS
    SLdefine_for_ifdef ("VMS");
+# else
+   SLdefine_for_ifdef ("UNIX");
+# endif
+#endif
+
+#if 0
+   if (NULL != getenv ("AUTOSUBSCRIBE"))
+     Slrn_Unsubscribe_New_Groups = 0;
+   if (NULL != getenv ("AUTOUNSUBSCRIBE"))
+     Slrn_Unsubscribe_New_Groups = 1;
 #endif
    
    for (i = 1; i < argc; i++)
      {
-	if (!strcmp ("-create", argv[i])) create_flag = 1;
+	if (!strcmp ("--spool", argv[i])
+	    || !strcmp ("--nntp", argv[i])
+	    || !strcmp ("--inews", argv[i]))
+	  {
+	     i += parse_object_args (argv[i] + 2, argv + (i + 1), argc - (i + 1));
+	  }
+	else if (!strcmp ("--help", argv[i])) usage (NULL);
+	else if (!strcmp ("-create", argv[i])) create_flag = 1;
 	else if (!strcmp ("-C", argv[i])) use_color = 1;
-	else if (!strcmp ("-a", argv[i])) Slrn_List_Active_File = 1;
+	else if (!strcmp ("-a", argv[i])) use_active = 1;
 	else if (!strcmp ("-n", argv[i])) no_new_groups = 1;
 	else if (!strcmp ("-d", argv[i])) dsc_flag = 1;
 	else if (!strcmp ("-m", argv[i])) use_mouse = 1;
 	else if (!strcmp ("-k", argv[i])) no_score_file = 1;
 	else if (!strcmp ("-k0", argv[i]))
 	  Slrn_Perform_Scoring &= ~SLRN_EXPENSIVE_SCORING;
-	
+	else if (!strcmp ("--version", argv[i]))
+	  {
+	     version ();
+	  }
 	else if (!strncmp ("-D", argv[i], 2) && (argv[i][2] != 0))
 	  {
 	     if (SLdefine_for_ifdef (argv[i] + 2) == 0)
@@ -494,18 +792,37 @@ int main (int argc, char **argv)
 	  }
 	else if (i + 1 < argc)
 	  {
-	     if (!strcmp("-h", argv[i])) host = argv[++i];
-	     else if (!strcmp ("-f", argv[i])) Slrn_Newsrc_File = argv[++i];
+	     if (!strcmp ("-f", argv[i])) Slrn_Newsrc_File = argv[++i];
 	     else if (!strcmp ("-i", argv[i])) init_file = argv[++i];
-	     else if (!strcmp ("-p", argv[i]))
+	     else 
 	       {
-		  Slrn_NNTP_Port = atoi (argv[++i]);
+		  i += parse_object_args (NULL, argv + i, argc - i);
+		  i -= 1;
 	       }
-	     else usage (NULL);
 	  }
-	else usage (NULL);
+	else 
+	  {
+	     i += parse_object_args (NULL, argv + i, argc - i);
+	     i -= 1;
+	  }
+     }
+   
+
+   if (-1 == slrn_init_objects ())
+     {
+	slrn_exit_error ("Error configuring server objects.");
      }
 
+
+   if (Slrn_Server_Id == 0) Slrn_Server_Id = SLRN_DEFAULT_SERVER_OBJ;
+   if (Slrn_Post_Id == 0) Slrn_Post_Id = SLRN_DEFAULT_POST_OBJ;
+   
+   if ((-1 == slrn_select_server_object (Slrn_Server_Id))
+       || (-1 == slrn_select_post_object (Slrn_Post_Id)))
+     {
+	slrn_exit_error ("Unable to select object.");
+     }
+     
    fprintf (stdout, "slrn %s\n", Slrn_Version);
 
    if (dsc_flag && create_flag)
@@ -513,26 +830,9 @@ int main (int argc, char **argv)
 	usage ("The -d and -create flags must not be specified together.");
      }
    
-   if ((host == NULL) && (NULL == (host = getserverbyfile(NNTPSERVER_FILE))))
-     {
-	fprintf (stderr, "You need to set the NNTPSERVER environment variable to your server name.\r\n");
-#ifdef VMS
-	fprintf (stderr, "Example: $ define/job NNTPSERVER my.news.server\r\n");
-#else
-# ifdef __os2__
-	fprintf (stderr, "Example: set NNTPSERVER=my.news.server\r\n");
-# else
-	fprintf (stderr, "Example (csh): setenv NNTPSERVER my.news.server\r\n");
-# endif
-#endif
-	exit (-1);
-     }
-   
-   Slrn_Server_Name = port_change (host);
-   
    SLtt_get_terminfo();
    if (use_color) SLtt_Use_Ansi_Colors = 1;
-   slrn_initialize ();
+   slrn_startup_initialize ();
    slrn_get_user_info ();
    
    if (Slrn_Lib_Dir == NULL)
@@ -550,6 +850,12 @@ int main (int argc, char **argv)
    slrn_read_startup_file (file);      /* global file for all users */
    slrn_read_startup_file (init_file);
    
+#ifndef __os2__
+   /* Allow blink characters if in mono */
+   if (SLtt_Use_Ansi_Colors == 0)
+     SLtt_Blink_Mode = 1;
+#endif
+   
    /* Now that we have read in the startup file, check to see if the user
     * has a username and a usable hostname.  Without those, we are not
     * starting up.
@@ -558,7 +864,8 @@ int main (int argc, char **argv)
      {
 	slrn_exit_error ("\
 Unable to find a fully qualified hostname.  You will have to specify a\r\n\
-hostname in your .slrnrc file (slrn.rc on VMS).\r\n");
+hostname in your %s file.\r\n", 
+			 SLRN_USER_SLRNRC_FILENAME);
      }
    if ((NULL == Slrn_User_Info.username)
        || (0 == *Slrn_User_Info.username)
@@ -582,24 +889,25 @@ cannot be constructed.  Try setting the USER environment variable.\r\n");
      }
    
    if ((Slrn_Newsrc_File == NULL)
-       && ((Slrn_Newsrc_File = slrn_map_file_to_host (host)) == NULL))
+       && ((Slrn_Newsrc_File = slrn_map_file_to_host (Slrn_Server_Obj->name)) == NULL))
 #if defined(VMS) || defined(__os2__)
      Slrn_Newsrc_File = "jnews.rc";
 #else
    Slrn_Newsrc_File = ".jnewsrc";
 #endif
 
+   if (use_active) Slrn_List_Active_File = 1;
    if (use_mouse) Slrn_Use_Mouse = -1; /* -1 forces it. */
    
    if (dsc_flag)
      {
-	if (nntp_initialize_server (Slrn_Server_Name) != 0)
+	if (Slrn_Server_Obj->initialize () != 0)
 	  {
 	     slrn_exit_error ("Unable to initialize server.");
 	     return (-1);
 	  }
 	slrn_get_group_descriptions ();
-	nntp_close_server ();
+	Slrn_Server_Obj->close ();
 	return 0;
      }
    
@@ -614,8 +922,8 @@ cannot be constructed.  Try setting the USER environment variable.\r\n");
 	     if (NULL == (fp = slrn_open_home_file (".newsrc", "r", file, 0)))
 	       {
 		  slrn_exit_error ("\r\nUnable to open %s.\r\n\
-If you want to create %s, use the command:\r\n\
-slrn -f %s -create\r\n", file, file, file);
+If you want to create %s, add command line options:\r\n\
+   -f %s -create\r\n", file, file, file);
 	       }
 	  }
 	slrn_fclose (fp);
@@ -642,9 +950,10 @@ slrn -f %s -create\r\n", file, file, file);
 
 void slrn_quit (int retcode)
 {
-   nntp_close_server ();
+   Slrn_Server_Obj->close ();
    slrn_reset_display (0);
    lock_file (0);
+   if (retcode) fprintf (stderr, "slrn: quiting on signal %d.\n", retcode);
    exit (retcode);
 }
 
@@ -653,16 +962,20 @@ void slrn_exit_error (char *fmt, ...)
 {
    va_list ap;
    
+   if (Slrn_Groups_Dirty) slrn_write_newsrc ();
+   
    if (Slrn_TT_Initialized)
      {
-	if (Slrn_Groups_Dirty) slrn_write_newsrc ();
 	slrn_reset_display (0);
-	client_close_server ();
+	Slrn_Server_Obj->close ();
      }
    
-   va_start (ap, fmt);
-   vfprintf (stderr, fmt, ap);
-   va_end(ap);
+   if (fmt != NULL)
+     {
+	va_start (ap, fmt);
+	vfprintf (stderr, fmt, ap);
+	va_end(ap);
+     }
    putc ('\n', stderr);
    lock_file (0);
    exit (-1);
