@@ -1,4 +1,6 @@
-#define Slrn_Version "0.8.7.1"
+#ifndef _SLRN_SLRN_H_
+#define SLRN_VERSION "0.8.8.4 (BETA)"
+extern char *Slrn_Version;
 
 extern int Slrn_TT_Initialized;
 extern void slrn_exit_error (char *, ...);
@@ -10,19 +12,18 @@ extern void slrn_reset_display (int);
 extern void slrn_suspend (int);
 extern int  slrn_get_new_news (int, int);
 extern char *Slrn_Newsrc_File;
-extern void (*Slrn_Hangup_Hook) (void);
+extern void (*Slrn_Hangup_Hook) (int);
 extern void (*Slrn_Winch_Function) (void);
 extern int slrn_parse_helpfile (char *);
 extern void slrn_smg_refresh (void);
 extern void slrn_set_suspension (int);
 extern void slrn_enable_mouse (int);
-
+extern void slrn_init_hangup_signals (int);
 extern int Slrn_Suspension_Ok;
 extern int Slrn_Use_Mouse;
 
 extern char *Slrn_Lib_Dir;
 extern char *Slrn_Current_Group;
-extern char *Slrn_Server_Name;
 
 #define HEADER_COLOR	1
 #define GROUP_COLOR	2
@@ -38,7 +39,9 @@ extern char *Slrn_Server_Name;
 #define HIGH_SCORE_COLOR 12
 #define MENU_PRESS_COLOR 13
 #define HEADER_NUMBER_COLOR 14
+#define GROUP_DESCR_COLOR 15
 /* Colors below here are inverse video on B+W terminals. */
 #define STATUS_COLOR	20
 #define MENU_COLOR	21
 
+#endif				       /* _SLRN_SLRN_H_ */
