@@ -11,7 +11,8 @@
  * Lines are considered commented out if they start with a '%' character.
  */
 
-#include <config.h>
+#include "config.h"
+#include "features.h"
 
 #include <stdio.h>
 #ifdef HAVE_STDLIB_H
@@ -30,39 +31,12 @@
 #include "startup.h"
 #include "score.h"
 #include "uudecode.h"
-#ifdef HAS_MIME_SUPPORT
-#include "mime.h"
+#if SLRN_HAS_MIME
+# include "mime.h"
 #endif
+#include "server.h"
 
 
-#if SLANG_VERSION < 9924
-#define SLcmd_execute_string sl_cmd_execute_string
-#define SLcmd_Cmd_Table_Type sl_cmd_cmd_table_type
-#define SLcmd_Cmd_Type sl_cmd_cmd_type
-
-typedef struct sl_cmd_cmd_table_type
-{
-   struct sl_cmd_cmd_type *table;
-   int argc;
-   char *string_args[SLCMD_MAX_ARGS];
-   int int_args[SLCMD_MAX_ARGS];
-   FLOAT float_args[SLCMD_MAX_ARGS];
-   int arg_type[SLCMD_MAX_ARGS];
-}
-sl_cmd_cmd_table_type;
-
-typedef struct sl_cmd_cmd_type
-{
-   int (*cmdfun)(int, struct sl_cmd_cmd_table_type *);
-   char cmd[32];
-   char arg_type[SLCMD_MAX_ARGS];
-}
-sl_cmd_cmd_type;
-
-static sl_cmd_execute_string (char *, SLcmd_Cmd_Table_Type *);
-extern int SLatoi(unsigned char *);
-#endif
-
 static SLcmd_Cmd_Table_Type Slrn_Cmd_Table;
 
 
@@ -70,6 +44,7 @@ static int unsetkey_fun (int, SLcmd_Cmd_Table_Type *);
 static int setkey_fun (int, SLcmd_Cmd_Table_Type *);
 static int server_fun (int, SLcmd_Cmd_Table_Type *);
 static int color_fun (int, SLcmd_Cmd_Table_Type *);
+static int mono_fun (int, SLcmd_Cmd_Table_Type *);
 static int user_data_fun (int, SLcmd_Cmd_Table_Type *);
 static int ignore_quote_fun (int, SLcmd_Cmd_Table_Type *);
 static int set_quote_fun (int, SLcmd_Cmd_Table_Type *);
@@ -83,6 +58,7 @@ static SLcmd_Cmd_Type Slrn_Startup_File_Cmds[] =
      {setkey_fun, "setkey", "SSS"},
      {server_fun, "server", "SS"},
      {color_fun, "color", "SSS"},
+     {mono_fun, "mono", "SSsss"},
      {set_variable_fun, "set", "SG"},
      {user_data_fun, "replyto", "S"},
      {user_data_fun, "organization", "S"},
@@ -93,30 +69,45 @@ static SLcmd_Cmd_Type Slrn_Startup_File_Cmds[] =
      {user_data_fun, "username", "S"},
      {user_data_fun, "scorefile", "S"},
      {user_data_fun, "cc_followup_string", "S"},
-#ifndef NO_DECODE
+#if SLRN_HAS_DECODE
      {user_data_fun, "decode_directory", "S"},
 #endif
      {user_data_fun, "editor_command", "S"},
      {nnrp_fun, "nnrpaccess", "SSS" },
-     {ignore_quote_fun, "ignore_quotes", "S"},
+#define SLRN_MAX_QUOTE_REGEXP 5
+     {ignore_quote_fun, "ignore_quotes", "Sssss"},
      {set_quote_fun, "quote_string", "S"},
      {autobaud_fun, "autobaud", ""},
      {NULL, "", ""}
 };
 
+SLRegexp_Type *Slrn_Ignore_Quote_Regexp [SLRN_MAX_QUOTE_REGEXP + 1];
+
 static int This_Line_Num;		       /* current line number in startup file */
 static char *This_File;
 static char *This_Line;		       /* line being parsed */
 
 static void exit_malloc_error (void)
 {
+   if (This_File == NULL)
+     slrn_exit_error ("Memory Allocation Failure");
+   
    slrn_exit_error ("%s: Line %d\n%sMemory Allocation Failure",
 		    This_File, This_Line_Num, This_Line);
 }
 
+static char *safe_malloc (unsigned int n)
+{
+   char *s;
+   s = (char *) SLMALLOC (n);
+   if (s == NULL) exit_malloc_error ();
+   return s;
+}
+
+
 static void exit_unknown_object (void)
 {
-   slrn_exit_error ("%s: Unsupported option on line %d\n%s",
+   slrn_exit_error ("%s: Error encountered processing line %d\n%s",
 		    This_File, This_Line_Num, This_Line);
 }
 
@@ -163,29 +154,57 @@ static int unsetkey_fun (int argc, SLcmd_Cmd_Table_Type *table)
    return 0;
 }
 
-static SLRegexp_Type Quote_Regexp;
-SLRegexp_Type *Slrn_Ignore_Quote_Regexp;
 
-static void compile_quote_regexp (char *str)
+static SLRegexp_Type *compile_quote_regexp (char *str)
 {
-   static unsigned char compiled_pattern_buf[256];
+   unsigned char *compiled_pattern_buf;
+   SLRegexp_Type *r;
    
-   Quote_Regexp.pat = (unsigned char *) str;
-   Quote_Regexp.buf = compiled_pattern_buf;
-   Quote_Regexp.case_sensitive = 1;
-   Quote_Regexp.buf_len = sizeof (compiled_pattern_buf);
+   compiled_pattern_buf = (unsigned char *) safe_malloc (512);
+   r = (SLRegexp_Type *) safe_malloc (sizeof (SLRegexp_Type));
    
-   if (SLang_regexp_compile (&Quote_Regexp))
+   r->pat = (unsigned char *) str;
+   r->buf = compiled_pattern_buf;
+   r->case_sensitive = 1;
+   r->buf_len = 512;
+   
+   if (SLang_regexp_compile (r))
      {
 	slrn_exit_error ("%s: line %d:\n%sInvalid regular expression.",
 			 This_File, This_Line_Num, This_Line);
      }
-   Slrn_Ignore_Quote_Regexp = &Quote_Regexp;
+   
+   return r;
 }
 
 static int ignore_quote_fun (int argc, SLcmd_Cmd_Table_Type *table)
 {
-   compile_quote_regexp (table->string_args[1]);
+   unsigned int i;
+   SLRegexp_Type *r;
+   
+   if (argc > SLRN_MAX_QUOTE_REGEXP + 1)
+     {
+	slrn_exit_error ("%s: line %d:\n%sToo many expressions specified.",
+			 This_File, This_Line_Num, This_Line);
+     }
+   
+   for (i = 0; i < SLRN_MAX_QUOTE_REGEXP; i++)
+     {
+	r = Slrn_Ignore_Quote_Regexp[i];
+	if (r != NULL)
+	  {
+	     if (r->buf != NULL)
+	       SLFREE (r->buf);
+	     SLFREE (r);
+	     Slrn_Ignore_Quote_Regexp [i] = NULL;
+	  }
+     }
+	
+   for (i = 1; i < argc; i++)
+     {
+	Slrn_Ignore_Quote_Regexp[i - 1] 
+	  = compile_quote_regexp (table->string_args[i]);
+     }
    return 0;
 }
 
@@ -208,6 +227,12 @@ Set_Int_Type;
 
 static Set_Int_Type Int_Things_To_Set [] =
 {
+     {"ignore_signature", &Slrn_Sig_Is_End_Of_Article},
+#if SLRN_HAS_NNTP_SUPPORT
+     {"query_reconnect", &Slrn_Query_Reconnect},
+#else
+     {"query_reconnect", NULL},
+#endif
      {"use_xgtitle", &Slrn_Use_Xgtitle},
      {"show_article", &Slrn_Startup_With_Article},
      {"author_display", &Slrn_Show_Author},
@@ -226,17 +251,32 @@ static Set_Int_Type Int_Things_To_Set [] =
      {"uncollapse_threads", &Slrn_Threads_Visible},
      {"read_active", &Slrn_List_Active_File},
      {"use_metamail", &Slrn_Use_Meta_Mail},
+     {"group_dsc_start_column", &Slrn_Group_Description_Column},
+     {"lines_per_update", &Slrn_Reads_Per_Update},
 #ifndef __os2__
      {"use_blink", &SLtt_Blink_Mode},
 #endif
      {"wrap_flags", &Slrn_Wrap_Mode},
+     {"write_newsrc_flags", &Slrn_Write_Newsrc_Flags},
      {"query_read_group_cutoff", &Slrn_Query_Group_Cutoff},
      {"prompt_next_group", &Slrn_Prompt_Next_Group},
      {"use_header_numbers", &Slrn_Use_Header_Numbers},
-#ifdef HAS_MIME_SUPPORT
+#if SLRN_HAS_SPOILERS
+     {"spoiler_char", &Slrn_Spoiler_Char},
+#else
+     {"spoiler_char", NULL},
+#endif
+#if SLRN_HAS_MIME
      {"use_mime", &Slrn_Use_Mime},
 #else
      {"use_mime", NULL},
+#endif
+#if 0
+#if SLRN_HAS_INEWS_SUPPORT
+     {"use_inews", &Slrn_Use_Inews},
+#else
+     {"use_inews", NULL},
+#endif
 #endif
      {NULL, NULL}
 };
@@ -250,15 +290,27 @@ Set_String_Type;
 
 static Set_String_Type String_Things_To_Set [] =
 {
+     {"non_Xbrowser", &Slrn_NonX_Browser},
+     {"Xbrowser", &Slrn_X_Browser},
      {"save_posts", &Slrn_Save_Posts_File},
      {"save_directory", &Slrn_Save_Directory},
-#ifndef NO_DECODE
-     {"decode_directory", &Slrn_Decode_Directory},
+     {"decode_directory", 
+#if SLRN_HAS_DECODE
+	  &Slrn_Decode_Directory
 #else
-     {"decode_directory", NULL},
+	  NULL
 #endif
+     },
+   
+     {"inews_program",
+#if SLRN_HAS_INEWS_SUPPORT && SLRN_HAS_USER_INEWS
+	  &Slrn_Inews_Pgm
+#else
+	  NULL
+#endif
+     },
      
-#ifdef HAS_MIME_SUPPORT
+#if SLRN_HAS_MIME
      {"mime_charset", &Slrn_Mime_Display_Charset},
 #else
      {"mime_charset", NULL},
@@ -266,6 +318,56 @@ static Set_String_Type String_Things_To_Set [] =
 #ifndef VMS
      {"sendmail_command", &Slrn_SendMail_Command},
 #endif
+     {"spool_inn_root", 
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Inn_Root
+#else
+     NULL
+#endif
+     },
+     {"spool_root",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Spool_Root
+#else
+     NULL
+#endif
+     },
+     {"spool_nov_root",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Nov_Root
+#else
+     NULL
+#endif
+     },
+     {"spool_nov_file",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Nov_File
+#else
+     NULL
+#endif
+     },
+     {"spool_active_file",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Active_File
+#else
+     NULL
+#endif
+     },
+     {"spool_activetimes_file",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_ActiveTimes_File
+#else
+     NULL
+#endif
+     },
+     {"spool_newsgroups_file",
+#if SLRN_HAS_SPOOL_SUPPORT
+     &Slrn_Newsgroups_File
+#else
+     NULL
+#endif
+     },
+   
      {NULL, NULL}
 };
 
@@ -323,7 +425,7 @@ typedef struct Server_List_Type
 }
 Server_List_Type;
 
-Server_List_Type *Server_List;
+static Server_List_Type *Server_List;
 
 static int server_fun (int argc, SLcmd_Cmd_Table_Type *table)
 {
@@ -334,11 +436,8 @@ static int server_fun (int argc, SLcmd_Cmd_Table_Type *table)
    
    n = strlen (the_file);
    
-   if ((NULL == (s = (Server_List_Type *) SLMALLOC (sizeof (Server_List_Type))))
-       || (file = SLMALLOC (n + 2 + strlen (the_host))) == NULL)
-     {
-	exit_malloc_error ();
-     }
+   s = (Server_List_Type *) safe_malloc (sizeof (Server_List_Type));
+   file = safe_malloc (n + 2 + strlen (the_host));
    
    strcpy (file, the_file);
    host = file + (n + 1);
@@ -376,6 +475,7 @@ Color_Handle_Type Color_Handles[] =
      {"thread_number", THREAD_NUM_COLOR},
      {"header_number", HEADER_NUMBER_COLOR},
      {"high_score", HIGH_SCORE_COLOR},
+     {"description", GROUP_DESCR_COLOR},
      {NULL, -1}
 };
 
@@ -385,6 +485,7 @@ static int color_fun (int argc, SLcmd_Cmd_Table_Type *table)
    char *what = table->string_args[1];
    char *fg = table->string_args[2];
    char *bg = table->string_args[3];
+   
    Color_Handle_Type *ct = Color_Handles;
    
    while (ct->name != NULL)
@@ -400,6 +501,50 @@ static int color_fun (int argc, SLcmd_Cmd_Table_Type *table)
    return 0;
 }
 
+#if SLANG_VERSION < 9932
+# ifdef __os2__
+void SLtt_set_mono (int a, char *b, SLtt_Char_Type c)
+{
+   (void) a;
+   (void) b;
+   (void) c;
+}
+# endif
+#endif
+
+static int mono_fun (int argc, SLcmd_Cmd_Table_Type *table)
+{
+   char *what = table->string_args[1];
+   char *attr;
+   int i;
+   
+   Color_Handle_Type *ct = Color_Handles;
+   
+   while (ct->name != NULL)
+     {
+	if (!strcmp (ct->name, what))
+	  {
+	     SLtt_Char_Type mono_attr = 0;
+	     for (i = 2; i < argc; i++)
+	       {
+		  attr = table->string_args[i];
+		  if (!strcmp (attr, "bold")) mono_attr |= SLTT_BOLD_MASK;
+		  else if (!strcmp (attr, "blink")) mono_attr |= SLTT_BLINK_MASK;
+		  else if (!strcmp (attr, "underline")) mono_attr |= SLTT_ULINE_MASK;
+		  else if (!strcmp (attr, "reverse")) mono_attr |= SLTT_REV_MASK;
+		  else if (!strcmp (attr, "none")) mono_attr = 0;
+		  else exit_unknown_object ();
+	       }
+	     SLtt_set_mono (ct->value, NULL, mono_attr);
+	     return 0;
+	  }
+	ct++;
+     }
+   exit_unknown_object ();
+   return 0;
+}
+
+
 /*----------------------------------------------------------------------*\
 * static int user_data_fun ();
  *
@@ -433,7 +578,7 @@ User_Info_Variable_Type User_Info_Variables[] =
      {"signature", &Slrn_User_Info.signature, 0},
      {"scorefile", &Slrn_Score_File, 0},
      {"cc_followup_string", &Slrn_Courtesy_CC_Message, 0},
-#ifndef NO_DECODE
+#if SLRN_HAS_DECODE
      {"decode_directory", &Slrn_Decode_Directory, 0},
 #endif
      {"editor_command", &Slrn_Editor, 0},
@@ -463,8 +608,7 @@ static int user_data_fun (int argc, SLcmd_Cmd_Table_Type *table)
 	     else
 	       {
 		  ptr = u->addr;
-		  if ((contents = SLMALLOC (n + 1)) == NULL)
-		    exit_malloc_error ();
+		  contents = safe_malloc (n + 1);
 		  strcpy (contents, field);
 		  
 		  if (*ptr != NULL) SLFREE (*ptr);
@@ -494,28 +638,30 @@ static int user_data_fun (int argc, SLcmd_Cmd_Table_Type *table)
 
 static int nnrp_fun (int argc, SLcmd_Cmd_Table_Type *table)
 {
+#if SLRN_HAS_NNTP_SUPPORT
    char *server = table->string_args[1];
    char *name = table->string_args[2];
    char *pass = table->string_args[3];
-   int n;
+   unsigned int len;
+   
+   if (Slrn_NNTP_Server_Name == NULL) 
+     return 0;
    
    /* Skip this one it is not the one that we are looking for */
-   if (strcmp(server, Slrn_Server_Name))
+   if (strcmp(server, Slrn_NNTP_Server_Name))
      return 0;
    
    if (Slrn_User_Info.nnrpname != NULL) SLFREE (Slrn_User_Info.nnrpname);
    if (Slrn_User_Info.nnrppass != NULL) SLFREE (Slrn_User_Info.nnrppass);
    
-   n = strlen(name);
-   if (NULL == (Slrn_User_Info.nnrpname = (char *) SLMALLOC(n + 1)))
-     exit_malloc_error ();
+   len = strlen(name);
+   Slrn_User_Info.nnrpname = safe_malloc (len + 1);
    strcpy(Slrn_User_Info.nnrpname, name);
    
-   n = strlen(pass);
-   if (NULL == (Slrn_User_Info.nnrppass = (char *) SLMALLOC(n + 1)))
-     exit_malloc_error ();
+   len = strlen(pass);
+   Slrn_User_Info.nnrppass = safe_malloc (len + 1);
    strcpy(Slrn_User_Info.nnrppass, pass);
-   
+#endif   
    return 0;
 }
 
@@ -525,36 +671,57 @@ static void slrn_init_keymaps (void)
    slrn_init_article_keymap ();
 }
 
-void slrn_initialize (void)
+void slrn_startup_initialize (void)
 {
    char *white = "white", *blue = "blue", *red = "red", *yellow = "yellow";
    char *black = "black";
    slrn_init_keymaps ();
    SLang_init_case_tables ();
-   compile_quote_regexp ("^ ?[:>]");
+   
+   Slrn_Ignore_Quote_Regexp[0] = compile_quote_regexp ("^ ?[:>]");
    
    /* default colors -- suitable for a color xterm */
    
    SLtt_set_color (0, NULL, black, white);
    SLtt_set_color (STATUS_COLOR, NULL, yellow, blue);
+   SLtt_set_mono (STATUS_COLOR, NULL, SLTT_REV_MASK);
    SLtt_set_color (MENU_COLOR, NULL, yellow, blue);
+   SLtt_set_mono (MENU_COLOR, NULL, SLTT_REV_MASK);
    SLtt_set_color (HEADER_COLOR, NULL, "brightcyan", white);
+   SLtt_set_mono (HEADER_COLOR, NULL, SLTT_BOLD_MASK);
    SLtt_set_color (GROUP_COLOR, NULL, blue, white);
+   SLtt_set_mono (GROUP_COLOR, NULL, 0);
    SLtt_set_color (SUBJECT_COLOR, NULL, black, white);
+   SLtt_set_mono (SUBJECT_COLOR, NULL, 0);
    SLtt_set_color (AUTHOR_COLOR, NULL, "magenta", white);
+   SLtt_set_mono (AUTHOR_COLOR, NULL, 0);
    SLtt_set_color (ERROR_COLOR, NULL, red, white);
+   SLtt_set_mono (ERROR_COLOR, NULL, SLTT_BLINK_MASK);
    SLtt_set_color (CURSOR_COLOR, NULL, "brightgreen", white);
+   SLtt_set_mono (CURSOR_COLOR, NULL, SLTT_REV_MASK);
    SLtt_set_color (ARTICLE_COLOR, NULL, blue, white);
+   SLtt_set_mono (ARTICLE_COLOR, NULL, 0);
    SLtt_set_color (TREE_COLOR, NULL, red, white);
+   SLtt_set_mono (TREE_COLOR, NULL, 0);
    SLtt_set_color (QUOTE_COLOR, NULL, red, white);
+   SLtt_set_mono (QUOTE_COLOR, NULL, 0);
    SLtt_set_color (SIGNATURE_COLOR, NULL, red, white);
+   SLtt_set_mono (SIGNATURE_COLOR, NULL, 0);
    SLtt_set_color (THREAD_NUM_COLOR, NULL, blue, white);
+   SLtt_set_mono (THREAD_NUM_COLOR, NULL, SLTT_BOLD_MASK);
    SLtt_set_color (HIGH_SCORE_COLOR, NULL, red, white);
+   SLtt_set_mono (HIGH_SCORE_COLOR, NULL, SLTT_BOLD_MASK);
    SLtt_set_color (HEADER_NUMBER_COLOR, NULL, "green", white);
+   SLtt_set_mono (HEADER_NUMBER_COLOR, NULL, 0);
    SLtt_set_color (MENU_PRESS_COLOR, NULL, blue, yellow);
+   SLtt_set_mono (MENU_PRESS_COLOR, NULL, 0);
+   SLtt_set_color (GROUP_DESCR_COLOR, NULL, "magenta", white);
+   SLtt_set_mono (GROUP_DESCR_COLOR, NULL, 0);
+   SLtt_set_mono (HIGH_SCORE_COLOR, NULL, SLTT_BOLD_MASK);
+   
    
 #ifndef __os2__
-   /* We are not using the blink characters */
+   /* We are not using the blink characters (unless in mono) */
    SLtt_Blink_Mode = 0;
 #endif
 }
@@ -565,7 +732,13 @@ void slrn_read_startup_file (char *name)
    char file[256];
    char line[256];
    SLPreprocess_Type pt;
+
+   if (-1 == slrn_init_readline ())
+     {
+	slrn_exit_error ("Unable to initialize S-Lang readline library.");
+     }
    
+
    if (-1 == SLprep_open_prep (&pt))
      {
 	slrn_exit_error ("Error initializing S-Lang preprocessor.");
@@ -609,192 +782,3 @@ char *slrn_map_file_to_host (char *host)
 }
 
 
-#if SLANG_VERSION < 9924
-static SLcmd_Cmd_Type *sl_cmd_find_command (char *s, SLcmd_Cmd_Type *cmd)
-{
-   char *cmdstr;
-   register char chs = *s++, ch;
-   while (1)
-     {
-	cmdstr = cmd->cmd;
-	ch = *cmdstr++;
-	if (ch == 0) break;
-	if ((ch == chs) && !strcmp (s, cmdstr)) return cmd;
-	cmd++;
-     }
-   return NULL;
-}
-
-static int sl_cmd_execute_string (char *str, SLcmd_Cmd_Table_Type *table)
-{
-   char *s, *b = NULL, *arg_type, *last_str;
-   SLcmd_Cmd_Type *cmd;
-   char buf[256];
-   int token_present;
-   int i, ret = -1, len;
-   int argc;
-   
-   if ((0 == SLang_extract_token (&str, buf, 0))
-       || (*buf == '%'))
-     return 0;
-   if (SLang_Error) return -1;
-   
-   if (((len = strlen (buf)) >= 32)
-       || (NULL == (cmd = sl_cmd_find_command (buf, table->table))))
-     {
-	SLang_doerror ("Invalid command.");
-	return -1;
-     }
-   
-   
-   if (NULL == (s = (char *) SLMALLOC(1 + len)))
-     {
-	SLang_Error = SL_MALLOC_ERROR;
-	return -1;
-     }
-   strcpy (s, buf);
-   argc = 0;
-   table->string_args[argc++] = s;
-   
-   arg_type = cmd->arg_type;
-   
-   while (*arg_type)
-     {
-	int guess_type = 0;
-	
-	last_str = str;
-	token_present = SLang_extract_token (&str, buf, 0);
-	if (SLang_Error) goto error;
-	if (token_present)
-	  {
-	     b = buf;
-	     len = strlen (b);
-	     if ((*b == '"') && (len > 1))
-	       {
-		  b++;
-		  len -= 2;
-		  b[len] = 0;
-		  guess_type = STRING_TYPE;
-	       }
-	     else guess_type = SLang_guess_type (buf);
-	  }
-	
-	switch (*arg_type++)
-	  {
-	     /* variable argument number */
-	   case 'v':
-	     if (token_present == 0) break;
-	   case 'V':
-	     if (token_present == 0)
-	       {
-		  SLang_doerror ("Expecting argument.");
-		  goto error;
-	       }
-	     
-	     last_str = slrn_skip_whitespace (last_str);
-	     len = strlen (last_str);
-	     str = last_str + len;
-	     
-	     s = SLmake_nstring (last_str, len);
-	     if (s == NULL) goto error;
-	     table->arg_type[argc] = STRING_TYPE;
-	     table->string_args[argc++] = s;
-	     break;
-	     
-	   case 's':
-	     if (token_present == 0) break;
-	   case 'S':
-	     if (token_present == 0)
-	       {
-		  SLang_doerror ("Expecting string argument.");
-		  goto error;
-	       }
-	     
-	     s = SLmake_nstring (b, len);
-	     
-	     if (s == NULL) goto error;
-	     table->arg_type[argc] = STRING_TYPE;
-	     table->string_args[argc++] = s;
-	     break;
-	     
-	     /* integer argument */
-	   case 'i':
-	     if (token_present == 0) break;
-	   case 'I':
-	     if ((token_present == 0) || (INT_TYPE != guess_type))
-	       {
-		  SLang_doerror ("Expecting integer argument.");
-		  SLang_Error = TYPE_MISMATCH;
-		  goto error;
-	       }
-	     
-	     table->arg_type[argc] = INT_TYPE;
-	     table->int_args[argc++] = SLatoi((unsigned char *) buf);
-	     break;
-	     
-	     /* floating point arg */
-#ifdef FLOAT_TYPE
-	   case 'f':
-	     if (token_present == 0) break;
-	   case 'F':
-	     if ((token_present == 0) || (STRING_TYPE == guess_type))
-	       {
-		  SLang_doerror ("Expecting floating point argument.");
-		  SLang_Error = TYPE_MISMATCH;
-		  goto error;
-	       }
-	     table->arg_type[argc] = FLOAT_TYPE;
-	     table->float_args[argc++] = atof(buf);
-	     break;
-#endif
-	     /* Generic type */
-	   case 'g':
-	     if (token_present == 0) break;
-	   case 'G':
-	     if (token_present == 0)
-	       {
-		  SLang_doerror ("Expecting an argument.");
-		  SLang_Error = TYPE_MISMATCH;
-		  goto error;
-	       }
-	     
-	     switch (guess_type)
-	       {
-		case INT_TYPE:
-		  table->arg_type[argc] = INT_TYPE;
-		  table->int_args[argc++] = SLatoi((unsigned char *) buf);
-		  break;
-		  
-		case STRING_TYPE:
-		  s = SLmake_nstring (b, len);
-		  if (s == NULL) goto error;
-		  
-		  table->arg_type[argc] = STRING_TYPE;
-		  table->string_args[argc++] = s;
-		  break;
-#ifdef FLOAT_TYPE
-		case FLOAT_TYPE:
-		  table->arg_type[argc] = FLOAT_TYPE;
-		  table->float_args[argc++] = atof(buf);
-#endif
-	       }
-	     break;
-	  }
-     }
-   
-   /*                 call function */
-   ret = (*cmd->cmdfun)(argc, table);
-   
-   error:
-   for (i = 0; i < argc; i++)
-     {
-	if (NULL != table->string_args[i])
-	  {
-	     SLFREE (table->string_args[i]);
-	     table->string_args[i] = NULL;
-	  }
-     }
-   return ret;
-   
-}
-#endif
