@@ -1,6 +1,5 @@
 extern void slrn_read_startup_file (char *);
-extern void slrn_initialize (void);
+extern void slrn_startup_initialize (void);
 extern char *slrn_map_file_to_host (char *);
-
 extern int Slrn_Autobaud;
 
